"""
    This module contains the interface for fake data sources that store a predefined data set that the 'SeqClu'
    algorithm can request data to process from.
"""
from abc import ABC, abstractmethod
from typing import List, Union

from numpy import ndarray

from seqclupv.library.interfaces.data_source import IDataSource


class IFakeDataSource(IDataSource, ABC):

    @property
    @abstractmethod
    def data(self) -> List[Union[ndarray, list]]:
        """
            This property stores the data provided by the data source.

            :return: The data provided by the data source.
        """
        pass

    @property
    @abstractmethod
    def numClasses(self) -> int:
        """
            The number of classes that are present in the data provided by the data source.

            :return: An integer representing the number of classes that are
            present in the data provided by the data source.
        """
        pass



"""
	This module contains a class that is responsible for generating data from a sine curve that are passed to the
	'SeqClu' algorithm.
"""

import random
from typing import List, Union, Tuple

import numpy as np
from numpy import ndarray

from seqclupv.library.interfaces.data_generator import IDataGenerator


class CurveGenerator(IDataGenerator):

	# PROPERTIES #

	@property
	def error(self) -> float:
		"""
			This property stores the error factor that is used to add
			noise to the sine curve with a random number generator.

			:return: A float representing the error factor.
		"""
		return self._error

	@property
	def freq(self) -> Tuple[float, float]:
		"""
			This property stores the lower and upper bound used to draw a sample from a uniform distribution that
			will be used as the frequency.

			:return: Two floats representing the lower and upper bound used to draw a sample from a uniform distribution
			that will be used as the frequency.
		"""
		return self._freq

	@property
	def n(self) -> int:
		"""
			This property stores the number of sequences that should be generated.

			:return: An integer that represents the number of sequences that should be generated.
		"""
		return self._n

	@property
	def phase(self) -> int:
		"""
			This property stores the phase that is used to generate the sine curve.

			:return: An integer that represents the phase that is used to generate the sine curve.
		"""
		return self._phase

	@property
	def samplingRate(self) -> int:
		"""
			Defines how the rate at which samples are generated. The higher the value, the fewer samples are generated.

			:return: An integer representing the rate at which samples are generated.
		"""
		return self._samplingRate

	# CONSTRUCTOR #

	def __init__(self, n: int, freq: Tuple[float, float], samplingRate: int, error: float, phase: int) -> None:
		"""
			This method initializes the data generator that is applicable to the sine curve data set with the parameters
			that are required to generate the data set.

			:param n: The amount of samples that are generated.
			:param freq: The lower and upper bound of the distribution that generates the frequency of a sample.
			:param samplingRate: The rate at which samples are generated.
			:param error: The error factor used to add noise to samples.
			:param phase: The phase that is used to generate the sine curve.
		"""
		super().__init__()
		self._error = error
		self._freq = freq
		self._n = n
		self._phase = phase
		self._samplingRate = samplingRate

	# PUBLIC METHODS #

	def generateData(self) -> List[Union[ndarray, list]]:
		"""
			This method generates a data set that is used in the 'SeqClu' algorithm or one of its extensions as
			a list of sequences.

			:return: A data set that is used in the 'SeqClu' algorithm or one of its extensions as
			a list of sequences.
		"""
		trajectory = []
		for i in range(self.n):
			freq = random.uniform(self.freq[0], self.freq[1])
			line = np.arange(1, 101, self.samplingRate)
			error = [random.random() * self.error for _ in range(len(line))]
			l = np.sin((freq * line) + self.phase) + error
			trajectory.append(l)
		return trajectory



"""
    This module contains an implementation of the data structure that stores all relevant information about one of the
    clusters that are maintained as part of the 'SeqClu' algorithm.
"""

from typing import Dict, Optional, Set, Tuple

import numpy as np
from numpy import ndarray

from seqclupv.library.storage.prototype_frequencies import PrototypeFrequencyStore
from seqclupv.library.storage.prototypes import PrototypeStore
from seqclupv.library.interfaces.distance_measure import IDistanceMeasure

from seqclupv.library.utilities.hash_filter import isHashInKey
from seqclupv.library.utilities.hash_sequence import hashSequence


class ClusterStore:

    @property
    def averageDistance(self) -> float:
        """
            This property stores the average distance between any prototype and all other prototypes in the cluster.

            :return: The average distance between any prototype and all other prototypes in the cluster.
        """
        if self._averageDistance is None:
            self._averageDistance = self.averageSumOfDistances / (self.prototypes.numPrototypes - 1)
        return self._averageDistance

    @property
    def averageDistanceFromRepresentativeToNonRepresentative(self) -> float:
        """
            This property stores the average distance from representative prototypes to non-representative prototypes
            in the cluster.

            :return: The average distance from representative prototypes to non-representative prototypes
            in the cluster.
        """
        if self._averageDistanceFromRepresentativeToNonRepresentative is None:
            self._averageDistanceFromRepresentativeToNonRepresentative = \
                self._calculateAverageSumOfDistancesRepresentativeToNonRepresentative() \
                / self.prototypes.numOtherPrototypes
        return self._averageDistanceFromRepresentativeToNonRepresentative

    @property
    def averageRepresentativeness(self) -> float:
        """
            This property stores the average representativeness of prototypes in the cluster.

            :return: The average representativeness of prototypes in the cluster.
        """
        if self._averageRepresentativeness is None:
            self._averageRepresentativeness = self._calculateAverageRepresentativeness()
        return self._averageRepresentativeness

    @property
    def averageSumOfDistances(self) -> float:
        """
            This property stores the average sum of distances between all pairs of prototypes in the cluster.

            :return: The average sum of distances between all pairs of prototypes in the cluster.
        """
        if self._averageSumOfDistances is None:
            self._averageSumOfDistances = self._calculateAverageSumOfDistances(False)
        return self._averageSumOfDistances

    @property
    def distanceMeasure(self) -> IDistanceMeasure:
        """
            This property stores the distance measure that is used in the 'SeqClu' algorithm.

            :return: The distance measure that is used in the 'SeqClu' algorithm.
        """
        return self._distanceMeasure

    @property
    def distances(self) -> Dict[Tuple[str, str], float]:
        """
            This property stores the pair-wise distances between pairs of (candidate) prototypes.

            :return: The pair-wise distances between pairs of (candidate) prototypes.
        """
        return self._distances

    @property
    def error(self) -> float:
        """
            This property stores the error that is made in every approximation of the distance from an incoming
            sequence to the cluster.

            :return: The error that is made in every approximation of the distance from an incoming
            sequence to the cluster.
        """
        if self._error is None:
            self._error = (1 - self.averageRepresentativeness) \
                          * self.averageDistanceFromRepresentativeToNonRepresentative
        return self._error

    @property
    def identifier(self) -> int:
        """
            This property stores the identifier of the cluster.

            :return: An integer representing the identifier of the cluster.
        """
        return self._identifier

    @property
    def prototypeFrequencies(self) -> PrototypeFrequencyStore:
        return self._prototypeFrequencies

    @property
    def prototypes(self) -> PrototypeStore:
        """
            This property stores the prototypes of the cluster.

            :return: The prototypes of the cluster.
        """
        return self._prototypes

    @property
    def sumsOfDistances(self) -> Dict[Tuple[str, bool], float]:
        """
            This property stores the sum of distances from a given prototype to all other prototypes in the cluster.

            :return: The sum of distances from a given prototype to all other prototypes in the cluster.
        """
        return self._sumsOfDistances

    @property
    def upperBound(self) -> float:
        """
            This property stores the upper bound of the distance between some incoming sequence and the cluster for the
            sequence to be considered as a candidate prototype.

            :return: The upper bound of the distance between some incoming sequence and the cluster for the
            sequence to be considered as a candidate prototype.
        """
        if self._upperBound is None:
            self._upperBound = self.averageDistance + self.error
        return self._upperBound

    def __init__(self, identifier: int, numRepresentativePrototypes: int, numPrototypes: int,
                 distanceMeasure: IDistanceMeasure, tick: int) -> None:
        """
            This method initializes the cluster given its identifier, the number of (representative) prototypes, the
            distance measure that should be used and the tick at which the cluster was initialized.

            :param identifier: The identifier of the cluster.
            :param numRepresentativePrototypes: The number of representative prototypes.
            :param numPrototypes: The number of prototypes.
            :param distanceMeasure: The distance measure that should be used by the cluster.
            :param tick: The tick at which the cluster was initialized.
        """
        self._averageDistance: Optional[float] = None
        self._averageDistanceFromRepresentativeToNonRepresentative: Optional[float] = None
        self._averageRepresentativeness: Optional[float] = None
        self._averageSumOfDistances: Optional[float] = None
        self._distanceMeasure = distanceMeasure
        self._distances = {}
        self._error = None
        self._identifier = identifier
        self._prototypeFrequencies = PrototypeFrequencyStore(numPrototypes)
        self._prototypes = PrototypeStore(numRepresentativePrototypes, numPrototypes, tick)
        self._sumsOfDistances = {}
        self._sumsOfRepresentativeDistances = {}
        self._upperBound: Optional[float] = None

    def computeAverageDistance(self, sequence: Tuple[Optional[str], Optional[ndarray]], representative: bool) -> float:
        """
            This method computes the average distance between some sequence and either the representative prototypes
            or all prototypes.

            :param sequence: The sequence for which the average distance between itself and some set of prototypes
            needs to be computed.
            :param representative: A boolean value indicating whether or not the distance to the representative
            prototypes should be computed. If this value is false, the distance to all prototypes is computed instead.
            :return: The average distance between the given sequence and either the representative prototypes or all
            prototypes.
        """
        sumOfDistances = self.sumOfDistancesOf(sequence, representative)
        if representative:
            return sumOfDistances / self.prototypes.numRepresentativePrototypes
        return sumOfDistances / self.prototypes.numPrototypes

    def isCandidate(self, sequence: Tuple[Optional[str], Optional[ndarray]],
                    minimumRepresentativeness: float, clusterAssignment: bool) -> Tuple[float, bool, bool]:
        """
            This method checks if a given sequence is a candidate prototype for the cluster.

            :param sequence: The sequence for which needs to be checked whether or not it is a candidate to become
            a prototype for the cluster.
            :param minimumRepresentativeness: The minimum representativeness that the prototypes should have in order
            for approximation of the distance to the cluster to be used. If the prototypes are not representative
            enough, the accurate distance to the cluster is computed instead.
            :param clusterAssignment: A boolean value indicating whether or not to approximate the distance to the
            cluster.
            :return: A tuple containing either the approximated or accurate distance from the sequence to the cluster,
            a boolean value indicating whether or not the sequence is a candidate prototype and whether or not the
            distance from the sequence to the cluster was approximated.
        """
        if clusterAssignment and self.isRepresentativeEnough(minimumRepresentativeness):
            approximatedDistance = self.computeAverageDistance(sequence, True)
            return approximatedDistance, approximatedDistance < self.upperBound, True
        else:
            accurateDistance = self.computeAverageDistance(sequence, False)
            return accurateDistance, accurateDistance < self.averageDistance, False

    def isRepresentativeEnough(self, minimumRepresentativeness: float) -> bool:
        """
            This method determines whether or not the representative prototypes of the cluster are representative
            enough.

            :param minimumRepresentativeness: The minimum average representativeness the representative prototypes
            should have for them to be considered representative enough.
            :return: A boolean value indicating whether or not the representative prototypes of the cluster are
            representative enough.
        """
        return self.averageRepresentativeness >= minimumRepresentativeness

    def pairwiseDistanceOf(self, sequenceOne: Tuple[Optional[str], Optional[ndarray]],
                           sequenceTwo: Tuple[Optional[str], Optional[ndarray]]) -> float:
        """
            This method calculates the distance of two sequences.
            The result of this computation is memoized and if the distance is requested again at a later point in time,
            the calculation is not done again to save time.

            :param sequenceOne: The first sequence for which the distance should be computed.
            :param sequenceTwo: The second sequence for which the distance should be computed.
            :return: The computed distance between the two sequences.
        """
        sequenceOneHash, _ = sequenceOne
        sequenceTwoHash, _ = sequenceTwo
        if sequenceOneHash is not None and sequenceTwoHash is not None:
            if sequenceOneHash == sequenceTwoHash:
                return 0
            if (sequenceOneHash, sequenceTwoHash) in self.distances:
                return self.distances[(sequenceOneHash, sequenceTwoHash)]
            if (sequenceTwoHash, sequenceOneHash) in self.distances:
                return self.distances[(sequenceTwoHash, sequenceOneHash)]

        sequenceOneArray: ndarray = self._getSequence(sequenceOne)
        sequenceTwoArray: ndarray = self._getSequence(sequenceTwo)

        result = self.distanceMeasure.calculateDistance(sequenceOneArray, sequenceTwoArray)
        # Memoize the computed result since it has not been computed before.
        if sequenceOneHash is None:
            sequenceOneHash = hashSequence(sequenceOneArray)
        if sequenceTwoHash is None:
            sequenceTwoHash = hashSequence(sequenceTwoArray)
        self.distances[(sequenceOneHash, sequenceTwoHash)] = result
        self.distances[(sequenceTwoHash, sequenceOneHash)] = result

        return result

    def processSequenceIndefinitely(self, sequenceHash: str) -> None:
        """
            This method processes a sequence indefinitely, which means it is no longer a (candidate) prototype, but
            merely a sequence. As a result, all the stored information related to this sequence needs to be removed
            from the data structures that make up the cluster. This involves removing all pair-wise distances involving
            the sequence and removing the sequence from the data structure storing the sums of distances.

            :param sequenceHash: The hash of the sequence that needs to be processed indefinitely.
            :return: void
        """
        distancesToRemove = list(filter(lambda dictKey: isHashInKey(dictKey, sequenceHash), self.distances.keys()))
        minKey = None
        minDistance = np.inf
        for key in distancesToRemove:
            hashOne, hashTwo = key
            if (hashOne == sequenceHash or hashTwo == sequenceHash) and \
                    (hashOne in self.prototypes.prototypes or hashTwo in self.prototypes.prototypes):
                distance = self.distances[key]
                if distance < minDistance:
                    minDistance = distance
                    minKey = key
        if minKey is not None:
            hashOne, hashTwo = minKey
            if hashOne != sequenceHash:
                self.prototypeFrequencies.closestPrototypeObserved(hashOne, 1)
            else:
                self.prototypeFrequencies.closestPrototypeObserved(hashTwo, 1)

        for key in distancesToRemove:
            del self.distances[key]

        if (sequenceHash, True) in self.sumsOfDistances:
            del self.sumsOfDistances[(sequenceHash, True)]

        if (sequenceHash, False) in self.sumsOfDistances:
            del self.sumsOfDistances[(sequenceHash, False)]

    def representativenessOfSequence(self, sequence: Tuple[Optional[str], Optional[ndarray]]) -> float:
        """
            This method calculates the representativeness of a sequence relative to the prototypes of the cluster.

            :param sequence: The sequence for which the representativeness should be calculated.
            :return: A float value between 0 and 1 representing the representativeness of the sequence relative to
            the prototypes of the cluster.
        """
        sumOfDistances = self.sumOfDistancesOf(sequence, False)
        result = self.averageSumOfDistances / (2 * sumOfDistances)
        return result

    def sumOfDistancesOf(self, sequence: Tuple[Optional[str], Optional[ndarray]],
                         representative: bool) -> float:
        """
            This method computes the sum of distances between some sequence and either the representative prototypes
            or all prototypes.

            :param sequence: The sequence for which the sum of distances to either the representative prototypes
            or all prototypes needs to be computed.
            :param representative: A boolean value indicating whether or not the distance to the representative
            prototypes should be computed. If this value is false, the distance to all prototypes is computed instead.
            :return: The computed sum of distances. This sum of distances is also memoized such that it can be re-used
            at a later point in time.
        """
        sequenceHash, _ = sequence
        assert sequenceHash is not None
        if representative:
            if (sequenceHash, True) not in self.sumsOfDistances:
                sumOfDistances = self._sumOfDistancesOf(sequence, True, False)
                self.sumsOfDistances[(sequenceHash, True)] = sumOfDistances
            return self.sumsOfDistances[(sequenceHash, True)]

        # This function applies memoization to memoize the result of computations that were carried out previously.
        if (sequenceHash, False) not in self.sumsOfDistances:
            if (sequenceHash, True) in self.sumsOfDistances:
                otherSumOfDistances = self._sumOfDistancesOf(sequence, False, True)
                self.sumsOfDistances[(sequenceHash, False)] = self.sumsOfDistances[(sequenceHash,
                                                                                    True)] + otherSumOfDistances
            else:
                sumOfDistances = self._sumOfDistancesOf(sequence, False, False)
                # TODO: Also store sumsOfDistances[(sequenceHash, True)]?
                self.sumsOfDistances[(sequenceHash, False)] = sumOfDistances
        return self.sumsOfDistances[(sequenceHash, False)]

    def updatePrototypes(self, newPrototypes: Dict[str, ndarray], newOtherPrototypeHashes: Set[str],
                         newRepresentativePrototypeHashes: Set[str], tick: int) -> Set[str]:
        """
            This method updates the prototypes of the cluster.

            :param newPrototypes: The new prototypes as a dictionary where the keys are the hashes of the new prototypes
            and the values are the prototypes itself.
            :param newOtherPrototypeHashes: The hashes of new non-representative prototypes.
            :param newRepresentativePrototypeHashes: The hashes of new representative prototypes.
            :param tick: The tick at which the prototypes are updated.
            :return: The hashes of the prototypes that were removed.
        """
        assert len(self.prototypeFrequencies.frequencies) == self.prototypeFrequencies.numPrototypes
        newPrototypeHashes = set(newPrototypes.keys())
        oldPrototypeHashes = set(self.prototypeFrequencies.frequencies.keys())
        addedPrototypeHashes = newPrototypeHashes.difference(oldPrototypeHashes)
        removedPrototypeHashes = oldPrototypeHashes.difference(newPrototypeHashes)
        assert len(addedPrototypeHashes) == len(removedPrototypeHashes)

        self._computeRequiredDistances(newPrototypes, removedPrototypeHashes)
        self.prototypeFrequencies.updatePrototypes(newPrototypeHashes, addedPrototypeHashes,
                                                   removedPrototypeHashes, self.distances)

        result = self.prototypes.updatePrototypes(newPrototypes, newOtherPrototypeHashes,
                                                  newRepresentativePrototypeHashes, tick)
        if len(result) > 0:
            # TODO: We should discard only the discardable information here.
            self._averageSumOfDistances = None
            self._averageDistance = None
            self._upperBound = None
            self._error = None
            self._averageRepresentativeness = None
            self._averageDistanceFromRepresentativeToNonRepresentative = None
            self._sumsOfDistances = {}
            self._sumsOfRepresentativeDistances = {}

        return result

    # PRIVATE METHODS #

    def _calculateAverageRepresentativeness(self) -> float:
        """
            This method calculates the average representativeness of the representative prototypes of the cluster.

            :return: The average representativeness of the representative prototypes of the cluster.
        """
        result = 0
        for prototypeHash in self.prototypes.representativePrototypeHashes:
            result += self.representativenessOfSequence((prototypeHash, None))
        return result / self.prototypes.numRepresentativePrototypes

    def _calculateAverageSumOfDistances(self, representative: bool) -> float:
        """
            This method calculates the average sum of distances from any of the prototypes to all other prototypes
            in the cluster.

            :param representative: A boolean value indicating whether or not only the representative prototypes or all
            prototypes should be considered.
            :return: The computed average sum of distances for either the representative prototypes or all prototypes.
        """
        result = 0

        if representative:
            prototypeHashes = self.prototypes.representativePrototypeHashes
        else:
            prototypeHashes = self.prototypes.prototypes.keys()

        for prototypeHash in prototypeHashes:
            sumOfDistances = self.sumOfDistancesOf((prototypeHash, None), representative)
            result += sumOfDistances

        if representative:
            result /= self.prototypes.numRepresentativePrototypes
        else:
            result /= self.prototypes.numPrototypes

        return result

    def _calculateAverageSumOfDistancesRepresentativeToNonRepresentative(self) -> float:
        """
            This method calculates the average sum of distances from the representative prototypes to
            the non-representative prototypes.

            :return: A float representing the average sum of distances from the representative prototypes to
            the non-representative prototypes.
        """
        result = 0
        for representativePrototypeHash in self.prototypes.representativePrototypeHashes:
            result += self._sumOfDistancesOf((representativePrototypeHash, None), False, True)
        return result / self.prototypes.numRepresentativePrototypes

    def _computeRequiredDistances(self, newPrototypes: Dict[str, ndarray], removedPrototypeHashes: Set[str]) -> None:
        """
            This method computes the pair-wise distances that are required to update the prototypes of the cluster.
            These distances are memoized in the 'distances' data structure.

            :param newPrototypes: The new prototypes as a dictionary, where the keys are the prototype hashes and the
            values are the prototypes itself.
            :param removedPrototypeHashes: The set of hashes of the prototypes that were removed.
            :return: void
        """
        for newPrototypeHashOne, newPrototypeDataOne in newPrototypes.items():
            for removedPrototypeHash in removedPrototypeHashes:
                if (newPrototypeHashOne, removedPrototypeHash) not in self.distances \
                        and (removedPrototypeHash, newPrototypeHashOne) not in self.distances:
                    distance = self.distanceMeasure.calculateDistance(newPrototypeDataOne,
                                                                      self.prototypes.prototypes[removedPrototypeHash])
                    self.distances[(newPrototypeHashOne, removedPrototypeHash)] = distance
                    self.distances[(removedPrototypeHash, newPrototypeHashOne)] = distance
            for newPrototypeHashTwo, newPrototypeDataTwo in newPrototypes.items():
                if (newPrototypeHashOne, newPrototypeHashTwo) not in self.distances \
                        and (newPrototypeHashTwo, newPrototypeHashOne) not in self.distances:
                    distance = self.distanceMeasure.calculateDistance(newPrototypeDataOne, newPrototypeDataTwo)
                    self.distances[(newPrototypeHashOne, newPrototypeHashTwo)] = distance
                    self.distances[(newPrototypeHashTwo, newPrototypeHashOne)] = distance

    def _getSequence(self, sequence: Tuple[Optional[str], Optional[ndarray]]) -> ndarray:
        """
            This method retrieves the sequence as an array from a tuple containing an optional sequence hash and an
            optional sequence array.

            :param sequence: A tuple containing an optional sequence hash and an
            optional sequence array.
            :return: The sequence as an array.
        """
        sequenceHash, sequenceData = sequence
        if sequenceHash is not None and sequenceHash in self.prototypes.prototypes:
            return self.prototypes.getPrototype(sequenceHash)
        if sequenceData is not None:
            return sequenceData
        raise ValueError("The provided tuple of a sequence hash and an optional sequence cannot be traced to "
                         "a sequence.\nNOTE: This should never happen!")

    def _sumOfDistancesOf(self, sequence: Tuple[Optional[str], Optional[ndarray]],
                          representative: bool, onlyNonRepresentativePrototypes: bool) -> float:
        """
            This method computes the sum of distances between some sequence and either representative prototypes,
            non-representative prototypes or all prototypes.

            :param sequence: The sequence for which the sum of distances should be computed.
            :param representative: A boolean value indicating whether or not the sum of distances should be computed
            for the representative prototypes.
            :param onlyNonRepresentativePrototypes: A boolean value indicating whether or not the sum of distances
            should be computed only for the non-representative prototypes.
            :return: The sum of distances for the sequence and the given set of prototypes to compare to.
        """
        assert not (representative and onlyNonRepresentativePrototypes)
        result = 0
        # In this case, the sum of distances between the given sequence and all representative prototypes in the
        # cluster needs to be computed.
        if representative:
            comparePrototypeHashes = self.prototypes.representativePrototypeHashes
        # In this case, the sum of distances between the given sequence and all other prototypes in the cluster
        # needs to be computed.
        elif onlyNonRepresentativePrototypes:
            comparePrototypeHashes = self.prototypes.otherPrototypeHashes
        else:
            comparePrototypeHashes = self.prototypes.prototypes.keys()

        for comparePrototypeHash in comparePrototypeHashes:
            prototype = (comparePrototypeHash, None)
            result += self.pairwiseDistanceOf(sequence, prototype)

        # The result is a tuple of the sum of distances and an optional float containing the error in the computation.
        return result



"""
    This class can be used to process a stream of data in a separate thread while some other thread performs work.
    The data stream is clustered using the 'SeqClu' algorithm and the labels of items that have already been seen can
    be requested real-time.
"""

from copy import deepcopy
from typing import Dict, List, Optional, Set, Tuple, Union

import numpy as np
from numpy import ndarray

from seqclupv.library.interfaces.data_source import IDataSource
from seqclupv.library.interfaces.distance_measure import IDistanceMeasure
from seqclupv.library.interfaces.heuristics.prototype_value import IPrototypeValue
from seqclupv.library.interfaces.seqclu import ISeqClu
from seqclupv.library.storage.candidates import CandidateStore
from seqclupv.library.storage.cluster import ClusterStore
from seqclupv.library.storage.sequences import SequenceStore
from seqclupv.library.utilities.hash_sequence import hashSequence


class SeqClu(ISeqClu):

    # STATIC METHODS #

    @staticmethod
    def assignToCluster(clusters: List[ClusterStore], sequence: Tuple[Optional[str], Optional[ndarray]],
                        distances: List[Tuple[int, float, float]], clusterAssignment: bool) -> Tuple[int, bool]:
        """
            This method assigns a sequence to one of the provided clusters using a dictionary of pair-wise distances.

            :param clusters: The clusters that the sequence can be assigned to.
            :param sequence: The sequence that needs to be assigned to one of the clusters.
            :param distances: The pair-wise distances stored in a dictionary identified by a tuple of prototype hashes
            that identify the two prototypes.
            :param clusterAssignment: A boolean value indicating whether or not to approximate the distance to the
            cluster.
            :return: A tuple containing the index of the cluster that the sequence was assigned to and a boolean value
            indicating whether or not the distances used to make the cluster assignment decision were approximated.
        """
        distances.sort(key=lambda x: x[1])
        if not clusterAssignment:
            return distances[0][0], False
        assignedClusterIdx, distance, error = distances[0]
        result: Set[int] = set([])
        result.add(assignedClusterIdx)
        for i in range(1, len(distances)):
            otherClusterIdx, otherDistance, otherError = distances[i]
            if SeqClu.isAmbiguous((distance, error), (otherDistance, otherError)):
                result.add(otherClusterIdx)
        if len(result) == 1:
            return assignedClusterIdx, True
        ambiguousClusters = list(map(lambda x: clusters[x], list(result)))
        return SeqClu.assignToClusterAccurate(ambiguousClusters, sequence), False

    @staticmethod
    def assignToClusterAccurate(clusters: List[ClusterStore], sequence: Tuple[Optional[str], Optional[ndarray]]) -> int:
        """
            This method assigns a sequence to one of the provided clusters using accurate distances.

            :param clusters: The clusters that the sequence can be assigned to.
            :param sequence: The sequence that needs to be assigned to one of the clusters.
            :return: The index of the cluster that the sequence was assigned to.
        """
        label = -1
        minDistance = np.inf
        for cluster in clusters:
            distance = cluster.sumOfDistancesOf(sequence, False)
            if distance < minDistance:
                label = cluster.identifier
                minDistance = distance
        return label

    @staticmethod
    def computeDistanceToClusters(clusters: List[ClusterStore], sequence: Tuple[Optional[str], Optional[ndarray]],
                                  clusterAssignment: bool) -> List[Tuple[int, float, float]]:
        """
            This method computes the distance of a sequence to all clusters.

            :param clusters: The clusters for which the distance to the sequence needs to be computed.
            :param sequence: The sequence for which the distance to the clusters needs to be computed.
            :param clusterAssignment: A boolean value indicating whether or not to approximate the distance to the
            cluster.
            :return: A list of tuples containing the index of a cluster, the computed distance and the error made in
            this computation. For accurate distances, this error is 0.
        """
        result: List[Tuple[int, float, float]] = []
        for cluster in clusters:
            assert cluster.prototypes.fullyInitialized
            distance = cluster.computeAverageDistance(sequence, clusterAssignment)
            if clusterAssignment:
                error = cluster.error
            else:
                error = 0
            result.append((cluster.identifier, distance, error))
        return result

    @staticmethod
    def determineCandidacy(candidates: CandidateStore,
                           clusters: List[ClusterStore],
                           sequence: Tuple[Optional[str], Optional[ndarray]],
                           minimumRepresentativeness: float,
                           clusterAssignment: bool) \
            -> Tuple[List[Tuple[int, float, float]], Set[int]]:
        """
            This method determines for which clusters a given sequence is a candidate.

            :param candidates: The buffer of candidate prototypes.
            :param clusters: The clusters that the given sequence could become a prototype for.
            :param sequence: The sequence for which needs to be determined whether or not it is a candidate to become
            a prototype for the clusters.
            :param minimumRepresentativeness: The minimum representativeness that the prototypes should have in order
            for approximation of the distance to the cluster to be used. If the prototypes are not representative
            enough, the accurate distance to the cluster is computed instead.
            :param clusterAssignment: A boolean value indicating whether or not to approximate the distance to the
            cluster.
            :return: A tuple containing a list of tuples that each contain the index of some cluster, the computed
            distance and the error made in this computation, as well as the indices of the clusters that the sequence
            is a candidate for.
        """
        # This function returns a list of tuples containing the cluster identifier, an approximation of the average
        # distance from the incoming sequence to the cluster and the upper bound of the error of the approximation.
        result: List[Tuple[int, float, float]] = []
        candidateFor = set([])
        sequenceHash, _ = sequence
        for cluster in clusters:
            assert cluster.prototypes.fullyInitialized
            distance, candidacy, isApproximation = cluster.isCandidate(sequence, minimumRepresentativeness,
                                                                       clusterAssignment)
            if isApproximation:
                error = cluster.error
            else:
                error = 0
            result.append((cluster.identifier, distance, error))
            if sequenceHash in cluster.prototypes.prototypes or sequenceHash in candidates.candidates:
                continue
            if candidacy:
                candidateFor.add(cluster.identifier)
        return result, candidateFor

    @staticmethod
    def initializeClusters(numClusters: int, numRepresentativePrototypes: int,
                           numPrototypes: int, distanceMeasure: IDistanceMeasure,
                           tick: int) -> List[ClusterStore]:
        """
            This method initializes the clusters that are used in the algorithm.

            :param numClusters: The number of clusters that need to be initialized.
            :param numRepresentativePrototypes: The number of representative prototypes each cluster needs to have.
            :param numPrototypes: The number of prototypes each cluster needs to have.
            :param distanceMeasure: The distance measure that is used in the algorithm.
            :param tick: The tick at which the clusters are initialized.
            :return: The resulting list of initialized clusters.
        """
        result: List[ClusterStore] = []
        for identifier in range(numClusters):
            result.append(ClusterStore(identifier, numRepresentativePrototypes, numPrototypes, distanceMeasure, tick))
        return result

    @staticmethod
    def isAmbiguous(distanceAndErrorOne: Tuple[float, float],
                    distanceAndErrorTwo: Tuple[float, float]) -> bool:
        """
            This method determines whether or not two computed distances given their errors are relatively too close
            to each other and are therefore ambiguous.

            :param distanceAndErrorOne: The distance and error that resulted from some distance computation.
            :param distanceAndErrorTwo: The distance and error that resulted from another distance computation.
            :return: A boolean value indicating whether or not the pair of distance computations is ambiguous.
        """
        distanceOne, errorOne = distanceAndErrorOne
        distanceTwo, errorTwo = distanceAndErrorTwo
        difference = abs(distanceOne - distanceTwo)
        return difference <= max(errorOne, errorTwo)

    @staticmethod
    def processCandidatesForCluster(cluster: ClusterStore, candidates: List[Tuple[str, ndarray]],
                                    numPrototypes: int, numRepresentativePrototypes: int,
                                    prototypeValue: IPrototypeValue, tick: int) -> Set[str]:
        """
            This method processes the candidate prototypes for a given cluster. The method decides whether or not
            the candidates are actually of high enough quality to become a prototype and if so, promotes these
            sequences to prototypes.

            :param cluster: The cluster for which the candidate prototypes should be processed.
            :param candidates: The candidate prototypes that should be processed.
            :param numPrototypes: The number of prototypes of all clusters.
            :param numRepresentativePrototypes: The number of representative prototypes of all clusters.
            :param prototypeValue: The heuristic function that is used to compute the value of a prototype.
            :param tick: The tick at which the candidate prototypes were processed.
            :return: The hashes of the prototypes that were discarded after processing all candidate prototypes.
        """
        prototypeValues: List[Tuple[Union[int, str], float]] = []
        prototypes = cluster.prototypes.prototypes

        # This set of statements should calculate the pair-wise distances between candidates and current prototypes.
        for i, candidate in enumerate(candidates):
            representativeness = cluster.representativenessOfSequence(candidate)
            prototypeValues.append((i, prototypeValue.evaluate(representativeness, 0)))

        # This set of statements should calculate the pair-wise distances between prototypes and prototypes.
        for (prototypeHash, prototype) in prototypes.items():
            representativeness = cluster.representativenessOfSequence((prototypeHash, prototype))
            weight = cluster.prototypeFrequencies.getWeight(prototypeHash)
            prototypeValues.append((prototypeHash, prototypeValue.evaluate(representativeness, weight)))

        # Sort the list to retrieve the p most representative prototypes from the set of candidateStore and prototypes.
        prototypeValues.sort(key=lambda x: x[1])

        newPrototypes: Dict[str, ndarray] = dict(
            list(map(lambda x: candidates[x[0]] if type(x[0]) == int else (x[0], prototypes[x[0]]),
                     prototypeValues[-numPrototypes:])))
        newRepresentativePrototypeHashes: Set[str] = set(map(lambda x: candidates[x[0]][0] if type(x[0]
                                                                                                   ) == int else x[0],
                                                             prototypeValues[-numRepresentativePrototypes:]))
        newOtherPrototypeHashes: Set[str] = set(map(lambda x: candidates[x[0]][0] if type(x[0]) == int else x[0],
                                                    prototypeValues[-numPrototypes:
                                                                    -numRepresentativePrototypes]))
        return cluster.updatePrototypes(newPrototypes, newOtherPrototypeHashes,
                                        newRepresentativePrototypeHashes, tick)

    # PROPERTIES #

    @property
    def bufferedSequences(self) -> Set[str]:
        """
            This property stores the sequences that were buffered by the algorithm.

            :return: The set of sequences that were buffered by the algorithm.
        """
        return self._bufferedSequences

    @property
    def buffering(self) -> bool:
        """
            This property stores a boolean value indicating whether or not the buffering feature is enabled.

            :return: A boolean value indicating whether or not the buffering feature is enabled.
        """
        return self._buffering

    @property
    def bufferFull(self) -> bool:
        """
            This method is a property that returns a boolean indicating whether or not the set of candidateStore has
            reached the maximum allowed size and is therefore full.

            :return: A boolean indicating whether or not the set of candidateStore has reached
            the maximum allowed size and is therefore full.
        """
        return len(self.candidateStore.candidates) + len(self.sequenceStore.sequences) >= self.bufferSize

    @property
    def bufferSize(self) -> int:
        """
            This method is a property that returns the maximum size of the buffer of candidateStore and ambiguous
            sequenceStore.

            :return: The maximum size of the buffer of candidateStore and ambiguous sequenceStore.
        """
        return self._bufferSize

    @property
    def candidateStore(self) -> CandidateStore:
        """
            This property stores the buffer of candidate prototypes.

            :return: The buffer of candidate prototypes.
        """
        return self._candidates

    @property
    def classes(self) -> List[Union[chr, int]]:
        """
            Tis property stores the classes that are present in the data set that is processed by the algorithm.

            :return: The classes that are present in the data set that is processed by the algorithm.
        """
        return self._classes

    @property
    def clusterAssignment(self) -> bool:
        """
            This property stores a boolean value indicating whether or not approximation of the distance to clusters
            is enabled in the algorithm.

            :return: A boolean value indicating whether or not approximation of the distance to clusters
            is enabled in the algorithm.
        """
        return self._clusterAssignment

    @property
    def clusteredByApproximation(self) -> Set[str]:
        """
            This property stores the sequences that were clustered by approximation by the algorithm.

            :return: The set of sequences that were clustered by approximation by the algorithm.
        """
        return self._clusteredByApproximation

    @property
    def clusters(self) -> List[ClusterStore]:
        """
            This property stores the clusters that the algorithm needs to work with.

            :return: The clusters that the algorithm needs to work with.
        """
        return self._clusters

    @property
    def finalLabels(self) -> Dict[str, int]:
        """
            This property stores the correct labels as a dictionary where the keys are the hashes of the sequences
            and the values are the correct labels.

            :return: The correct labels as a dictionary where the keys are the hashes of the sequences
            and the values are the correct labels.
        """
        labelsCopied = self.labels.copy()
        for cluster in self.clusters:
            for prototypeHash in cluster.prototypes.prototypes:
                labelsCopied[prototypeHash] = self.classes[cluster.identifier]
        return labelsCopied

    @property
    def fullyInitialized(self) -> bool:
        """
            This property stores a boolean value indicating whether or not the prototypes of all clusters are fully
            initialized.

            :return: A boolean value indicating whether or not the prototypes of all clusters are fully
            initialized.
        """
        for cluster in self.clusters:
            if not cluster.prototypes.fullyInitialized:
                return False
        return True

    @property
    def labels(self) -> Dict[str, int]:
        """
            This property stores the predicted labels of all sequences except for those that are prototypes.

            :return: The predicted labels of all sequences except for those that are prototypes.
        """
        return self._labels

    @property
    def minimumRepresentativeness(self) -> float:
        """
            This property stores the minimum representativeness that the prototypes should have in order
            for approximation of the distance to the cluster to be used. If the prototypes are not representative
            enough, the accurate distance to the cluster is computed instead.

            :return: The minimum representativeness that the prototypes should have in order
            for approximation of the distance to the cluster to be used. If the prototypes are not representative
            enough, the accurate distance to the cluster is computed instead.
        """
        assert 0 <= self._minimumRepresentativeness <= 1
        return self._minimumRepresentativeness

    @property
    def numFullyProcessed(self) -> int:
        """
            This property stores the number of sequences that were fully processed by the algorithm.

            :return: An integer representing the number of sequences that were fully processed by the algorithm.
        """
        return self._numFullyProcessed

    @property
    def numRepresentativePrototypes(self) -> int:
        """
            This property stores the number of representative prototypes of all clusters that are used by the algorithm.

            :return: An integer representing the number of representative prototypes
            of all clusters that are used by the algorithm.
        """
        return self._numRepresentativePrototypes

    @property
    def prototypeValueHeuristic(self) -> IPrototypeValue:
        """
            This property stores the heuristic function that is used to compute the value of a prototype.

            :return: The heuristic function that is used to compute the value of a prototype.
        """
        return self._prototypeValueHeuristic

    @property
    def sequenceStore(self) -> SequenceStore:
        """
            This property stores the buffer of sequences that is currently not used in the algorithm, but could be used
            in the future to store sequences that are likely to be assigned incorrectly.

            :return: The buffer of sequences that is currently not used in the algorithm, but could be used
            in the future to store sequences that are likely to be assigned incorrectly.
        """
        return self._sequences

    # CONSTRUCTOR #

    def __init__(self, dataSource: IDataSource, distanceMeasure: IDistanceMeasure,
                 numClusters: int, numRepresentativePrototypes: int, numPrototypes: int,
                 bufferSize: int, minimumRepresentativeness: float, prototypeValueHeuristic: IPrototypeValue,
                 clusterAssignment: bool, buffering: bool) -> None:
        """
            This method initializes the 'SeqClu' algorithm with a data source, distance measure, number of clusters
            and number of prototypes.

            :param dataSource: The data source that should be used in the 'SeqClu' algorithm.
            :param distanceMeasure: The distance measure that should be used in the 'SeqClu' algorithm.
            :param numClusters: The number of clusters that should be used in the 'SeqClu' algorithm.
            :param numRepresentativePrototypes: The number of representative prototypes.
            :param numPrototypes: The number of prototypes that should be used in the 'SeqClu' algorithm.'
            :param bufferSize: The maximum size of the buffer.
            :param minimumRepresentativeness: The minimum representativeness that the prototypes should have in order
            for approximation of the distance to the cluster to be used. If the prototypes are not representative
            enough, the accurate distance to the cluster is computed instead.
            :param prototypeValueHeuristic: The heuristic function that is used to compute the value of a prototype.
            :param clusterAssignment: A boolean value indicating whether or not to approximate the distance to the
            cluster.
            :param buffering: A boolean value indicating whether or not the buffering feature should be used.
            :return: void
        """
        super().__init__(dataSource, distanceMeasure, numClusters, numPrototypes)
        self._bufferedSequences = set([])
        self._buffering = buffering
        self._bufferSize = bufferSize
        self._candidates = CandidateStore(numRepresentativePrototypes, numPrototypes, -1)
        self._classes = dataSource.classes
        self._clusterAssignment = clusterAssignment
        self._clusteredByApproximation = set([])
        self._sequences = SequenceStore(-1)
        self._clusters = SeqClu.initializeClusters(numClusters, numRepresentativePrototypes, numPrototypes,
                                                   distanceMeasure, -1)
        self._labels = {}
        self._minimumRepresentativeness = minimumRepresentativeness
        self._numClusters = numClusters
        self._numFullyProcessed = 0
        self._numPrototypes = numPrototypes
        self._numRepresentativePrototypes = numRepresentativePrototypes
        self._prototypeValueHeuristic = prototypeValueHeuristic

    # PUBLIC METHODS #

    def alreadyProcessed(self, sequenceHash: str) -> bool:
        """
            This method determines if a sequence, identified by its hash, was already processed in the past.

            :param sequenceHash: The hash of the sequence for which needs to be determined whether or not it was already
            processed in the past.
            :return: A boolean value indicating whether or not the sequence was already processed in the past.
        """
        # Check if the sequence is already stored as a prototype of any of the clusters or as a candidate.
        for cluster in self.clusters:
            if sequenceHash in cluster.prototypes.prototypes:
                return True
        return sequenceHash in self.candidateStore.candidates or sequenceHash in self.labels

    def execute(self) -> None:
        """
            This method executes the 'SeqClu' algorithm from start to finish. The main loop over the data is
            implemented in here.

            :return: void
        """
        iteration = 0
        while not self.finish:
            iteration += 1
            print(f"[SeqClu] Iteration {iteration}.")
            finish = self._advanceTick()
            if finish:
                print("[SeqClu] The algorithm finished executing. Processing the buffer and returning...")
                break
        if self.buffering:
            # This if statement should not be necessary due to the nature of this function assuming the buffer is empty.
            self.forceProcessBuffer(True, self.tick)

    def forceProcessBuffer(self, persist: bool, tick: int) -> List[ClusterStore]:
        """
            This method forcefully processes the buffer of candidate prototypes. The buffer is then empty again and can
            fill up over time in the ticks that are to come.

            :param persist: A boolean value indicating whether or not to persist this action. If this value is set to
            false, the resulting list of clusters is returned, but all other data structures remain the same as before
            executing this method.
            :param tick: The tick at which the buffer is forcefully processed.
            :return: The clusters that are obtained after forcefully processing the buffer.
        """
        # This function forcefully empties the candidateStore and assigns all sequenceStore in it to a cluster.
        if persist:
            clusters = self.clusters
            candidates = self.candidateStore
        else:
            clusters = deepcopy(self.clusters)
            candidates = deepcopy(self.candidateStore)

        self._processCandidates(clusters, candidates, self.numClusters, self.numPrototypes,
                                self.numRepresentativePrototypes, tick)

        return clusters

    def processSequence(self, sequence: Tuple[str, ndarray], considerCandidacy: bool) -> None:
        """
            This method processes one sequence of the data set that is processed in the 'SeqClu' algorithm.

            :param sequence: The sequence that should be processed in the 'SeqClu' algorithm.
            :param considerCandidacy: A boolean value indicating whether or not the algorithm should consider if the
            sequence could be a prototype for any of the clusters.
            :return: void
        """
        if self.alreadyProcessed(sequence[0]):
            return
        if not self.fullyInitialized:
            for cluster in self.clusters:
                if not cluster.prototypes.fullyInitialized:
                    if not cluster.prototypes.representativePrototypesInitialized:
                        cluster.prototypes.addPrototype(sequence, True, self.tick, None)
                        cluster.prototypeFrequencies.initializePrototype(sequence[0])
                        return
                    else:
                        cluster.prototypes.addPrototype(sequence, False, self.tick, None)
                        cluster.prototypeFrequencies.initializePrototype(sequence[0])
                        return
        else:
            if considerCandidacy:
                averageDistances, candidateFor = SeqClu.determineCandidacy(self.candidateStore,
                                                                           self.clusters, sequence,
                                                                           self.minimumRepresentativeness,
                                                                           self.clusterAssignment)
                print(f"[SeqClu] Candidacy determined {list(candidateFor)}.")

                # If the incoming sequence is a candidate for any of the clusters, add the sequence to the buffer of candidateStore.
                if len(candidateFor) > 0:
                    if self.buffering:
                        sequenceHash, sequenceData = sequence
                        if sequenceHash is None:
                            sequenceHash = hashSequence(sequenceData)
                        self.bufferedSequences.add(sequenceHash)
                    self.candidateStore.addToCandidates(sequence, candidateFor, self.tick)
                    if self.bufferFull or not self.buffering:
                        print(f"[SeqClu] Forcefully emptying buffer...")
                        self.forceProcessBuffer(True, self.tick)
                    return
            else:
                averageDistances = SeqClu.computeDistanceToClusters(self.clusters, sequence, self.clusterAssignment)

            self._labelSequence(self.clusters, sequence, averageDistances)

    # PRIVATE METHODS #

    def _labelSequence(self, clusters: List[ClusterStore], sequence: Tuple[Optional[str], Optional[ndarray]],
                       averageDistances: List[Tuple[int, float, float]]) -> None:
        """
            This method labels a sequence based on the information that is available.

            :param clusters: The clusters at the given tick.
            :param sequence: The sequence that needs to be labeled.
            :param averageDistances: The average distance from the sequence to all clusters, represented as a list of
            tuples containing the index of some cluster, the average distance to that cluster and the error made in the
            distance computation for that cluster.
            :return: void
        """
        sequenceHash, sequenceData = sequence
        assignedCluster, byApproximation = SeqClu.assignToCluster(clusters, sequence,
                                                                  averageDistances, self.clusterAssignment)
        if byApproximation:
            if sequenceHash is None:
                sequenceHash = hashSequence(sequenceData)
            self.clusteredByApproximation.add(sequenceHash)
        self.labels[sequenceHash] = self.classes[assignedCluster]
        # TODO: Shouldn't this be called for all cluster instances?
        self.clusters[assignedCluster].processSequenceIndefinitely(sequenceHash)
        self._numFullyProcessed += 1

    def _processCandidates(self, clusters: List[ClusterStore], candidateStore: CandidateStore, numClusters: int,
                           numPrototypes: int, numRepresentativePrototypes: int, tick: int) -> None:
        """
            This method processes all the candidate prototypes in the buffer of candidate prototypes. This method is
            called when the buffer is forcefully processed.

            :param clusters: The clusters at the given tick.
            :param candidateStore: The buffer of candidate prototypes at the given tick.
            :param numClusters: The number of clusters that should be used in the 'SeqClu' algorithm.
            :param numPrototypes: The number of prototypes that should be used in the 'SeqClu' algorithm.
            :param numRepresentativePrototypes: The number of representative prototypes that should be used in the
            'SeqClu' algorithm.
            :param tick: The tick at which the buffer is forcefully processed.
            :return: The clusters that are obtained after forcefully processing the buffer.
        """
        for clusterIdx in range(numClusters):
            candidatesForCluster: List[Tuple[str, ndarray]] = list(map(lambda item: (item[0], item[1][0]),
                                                                       filter(lambda item: clusterIdx in item[1][1],
                                                                              candidateStore.candidates.items())))
            if len(candidatesForCluster) > 0:
                removedPrototypeHashes = SeqClu.processCandidatesForCluster(clusters[clusterIdx], candidatesForCluster,
                                                                            numPrototypes, numRepresentativePrototypes,
                                                                            self.prototypeValueHeuristic, tick)
                # Assumption that the removed prototype is still assigned to the cluster.
                for removedPrototypeHash in removedPrototypeHashes:
                    self.labels[removedPrototypeHash] = self.classes[clusterIdx]
                # TODO: Break should be added here, first fix and then verify if the performance remains the same.
        # All candidates have been processed, therefore we should empty the buffer.
        for candidateHash in list(candidateStore.candidates.keys()):
            isCandidate = False
            for cluster in clusters:
                # We assume here that candidates are never assigned to two clusters. TODO: Maybe ensure this?
                if candidateHash in cluster.prototypes.prototypes:
                    self.labels[candidateHash] = self.classes[cluster.identifier]
                    isCandidate = True
                    break
            if not isCandidate:
                candidate = (candidateHash, candidateStore.candidates[candidateHash][0])
                averageDistances = SeqClu.computeDistanceToClusters(clusters, candidate, self.clusterAssignment)
                self._labelSequence(clusters, candidate, averageDistances)
            candidateStore.removeFromCandidates(candidateHash)



"""
    This package is the top-level package of 'SeqClu'.
"""

from .library import HandwrittenCharacterGenerator, GesturePebbleGenerator, PLAIDGenerator, CurveGenerator, \
    FakeDataSource, DynamicTimeWarping, BasicBaselinePrototypes, BasicEvaluator, LinearPrototypeValue, \
    IBaselinePrototypes, ICandidateStore, IDataGenerator, IDataSource, IDistanceMeasure, IEvaluator, IFakeDataSource, \
    IPrototypeValue, ISeqClu, CandidateStore, ClusterStore, PrototypeFrequencyStore, PrototypeStore, SequenceStore, \
    calculate_f1_score, constructStream, isHashInKey, hashSequence, statistical_test, Visualizer
from .SeqClu import SeqClu
from .SeqCluBaselineOffline import SeqCluBaselineOffline
from .SeqCluBaselineOnline import SeqCluBaselineOnline

__all__ = ["HandwrittenCharacterGenerator",
           "GesturePebbleGenerator",
           "PLAIDGenerator",
           "CurveGenerator",
           "FakeDataSource",
           "DynamicTimeWarping",
           "BasicBaselinePrototypes",
           "BasicEvaluator",
           "LinearPrototypeValue",
           "IBaselinePrototypes",
           "ICandidateStore",
           "IDataGenerator",
           "IDataSource",
           "IDistanceMeasure",
           "IEvaluator",
           "IFakeDataSource",
           "IPrototypeValue",
           "ISeqClu",
           "CandidateStore",
           "ClusterStore",
           "PrototypeFrequencyStore",
           "PrototypeStore",
           "SequenceStore",
           "calculate_f1_score",
           "constructStream",
           "isHashInKey",
           "hashSequence",
           "statistical_test",
           "Visualizer",
           "main",
           "SeqClu",
           "SeqCluBaselineOffline",
           "SeqCluBaselineOnline"]



"""
    This package contains interfaces for heuristic functions that are used in variants of the 'SeqClu' algorithm.
"""

from .prototype_value import IPrototypeValue


__all__ = ["IPrototypeValue"]



"""
    This package contains all implementations of data generators that are used to generate data that are passed into
    the 'SeqClu' algorithm.
"""

from .handwritten_character_generator import HandwrittenCharacterGenerator
from .pebble_generator import GesturePebbleGenerator
from .plaid_generator import PLAIDGenerator
from .sine_curve_generator import CurveGenerator


__all__ = ["HandwrittenCharacterGenerator",
           "GesturePebbleGenerator",
           "PLAIDGenerator",
           "CurveGenerator"]



"""
    This module contains the class that stores the prototypes of some baseline. This class can be used
    to compare the prototypes obtained at the end of the online 'Seqclu' algorithm, where the number of identical
    prototypes divided by the total number of prototypes can be used as evaluation metric.
"""

from typing import List, Optional

from seqclupv.library.interfaces.baseline_prototypes import IBaselinePrototypes


class BasicBaselinePrototypes(IBaselinePrototypes):

    @staticmethod
    def getPrototypes(experimentName: str) -> Optional[List[List[str]]]:
        """
            This method returns the prototypes that were obtained after executing some baseline algorithm.
            These prototypes can be compared to the prototypes obtained at the end of the online 'SeqClu' algorithm,
            where the number of identical prototypes divided by the total number of prototypes
            can be used as evaluation metric.

            :param experimentName: The name of the experiment for which the prototypes need to be retrieved.
            :return: The prototypes obtained after executing some experiment identified by the experiment name.
        """
        if experimentName == "o29":
            return [['6756bc76', 'c64b021a', '5ac133ba', '7420af12', 'ff24fe7c', '2a64efc4', '6756bc76', '9c479a03'],
                    ['24acc8e9', '93a1ddc9', '482fed89', '89234e21', '482fed89', '76c30e17', '472b7fc1', '1332e09f'],
                    ['f418fa11', '4ccb3dd9', 'fa61df8d', '8ac4d5be', '69a451fd', '0dd66cc3', '69a451fd', '52eafac6']]
        if experimentName == "plaid":
            return None
        if experimentName == "pebbleTrain":
            return [['98c4a0e6', '237257d8', '4a9bf0ec', 'b81e204e', 'c456b122', '9fb22e3e'],
                    ['cfad21a6', 'fb45fb48', '930a101b', 'ce12103c', 'fff19865', '71ac95d0'],
                    ['1acf5df8', '98ebccea', '96dc73ce', '217f530d', '84ea0475', '8ba68fda'],
                    ['6b15001d', 'ac6f77f6', '93dcdf97', '638bb9d1', '2d0a0851', '6b50cf7b'],
                    ['fefa5265', '91a94be5', 'ee3b07c1', 'fa18a9a2', '00687b90', '2fba8027'],
                    ['c5b27ba8', '96b7bca0', 'a321ebc1', '59fed255', 'd725f213', '34a33aae']]
        if experimentName == "pebbleFull":
            return [['98ebccea', '2b4dcbd7', '96dc73ce', '0e6736c8', '84ea0475', '8ba68fda', '1acf5df8', '2b8eddc3'],
                    ['55ec131f', 'fb45fb48', 'eb219134', '03e1ae20', 'd524d948', '7db3fc75', '71ac95d0', 'cfad21a6'],
                    ['66f79f6b', '37b9d58e', '9497aaca', 'b81e204e', 'edf82d6b', 'ef47c484', 'ce0adac4', '886b30b1'],
                    ['6b15001d', 'ac6f77f6', '93dcdf97', '4f4bef57', '8211b5a0', '6cc481dc', '390acf90', 'af4c46ee'],
                    ['c794f77a', '2d21381c', 'f2b940ba', 'c4fc00b2', '919ee3a0', '54051a09', '4ea28372', 'c7b67231'],
                    ['3ee65e8f', '34107bcd', '8a5d1fe9', '6b50888c', 'c607d911', 'b92298a7', 'fa18a9a2', '906d7bc4']]
        if experimentName == "o295w":
            return [['6756bc76', 'ff24fe7c', '5ac133ba', '7420af12', '7e1a26ed', '2a64efc4', '9c479a03', 'c64b021a'],
                    ['482fed89', '7b8a937e', '76c30e17', '472b7fc1', '1a525726', '24acc8e9', '1332e09f', '93a1ddc9'],
                    ['69a451fd', '2e06ec79', '52eafac6', '6e6e22ef', 'fa61df8d', '4ccb3dd9', 'dfec2275', '0dd66cc3'],
                    ['8ac4d5be', '6b2cde41', '2ef44daf', '3ead38b7', '3f0aa84b', '0b371e7d', 'ba6e7c4a', 'f418fa11'],
                    ['89a684ef', '78ae25b3', '7a2e4832', '449dbb2e', '4f8ec214', '7c4cd0d9', '4cd6f30f', '349725c2']]
        if experimentName == "charFull":
            return None
        return None



"""
    This module contains a class that bundles several approaches to visualize the results of the variations of
    the 'SeqClu' algorithm that are contained in the package.

    NOTE: This class has actually never been used during the research project and therefore needs major modifications
    to make it compatible with the rest of the framework.
"""

import time

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from IPython import display
import pandas as pd
import seaborn as sns
from sklearn.manifold import TSNE

matplotlib.use("TkAgg")


class Visualizer:

    def __init__(self, classes, distribution, labels, indices, result, data, classDictionary,
                 numPrototypes, numClusters) -> None:
        self.classes = classes
        self.distribution = distribution
        self.labels = labels
        self.numPrototypes = numPrototypes
        self.numClasses = numClusters
        self.indices = indices
        self.result = result
        self.data = data
        self.classDictionary = classDictionary

    def visualizeInputData(self) -> None:
        """
            This method visualizes the input data in two dimensions.

            :return: void
        """
        fig = plt.figure(figsize=(10, 10))
        plt.title('Raw data')
        X_embedded = TSNE(random_state=42, n_components=2).fit_transform(self.distribution)
        # plt.scatter(*X_embedded)
        pal = sns.color_palette("hls", self.numClasses)  # 3 classes, hence 3 colors
        for i, txt in enumerate(self.labels):
            plt.scatter(X_embedded.T[0][i], X_embedded.T[1][i], color=pal[self.classDictionary[txt]])
            plt.annotate(i, (X_embedded.T[0][i], X_embedded.T[1][i]), color=pal[self.classDictionary[txt]], alpha=0.2)
        # Color = class, annotation = Sequence ID
        plt.show()

    def visualizeClustersAsTSNE(self) -> None:
        """
            This method visualizes the clusters as TSNE-graphs.

            :return: void
        """
        fig = plt.figure(figsize=(10, 10))
        plt.title('Clustered data')
        X_embedded = TSNE(random_state=42, n_components=2).fit_transform(self.distribution)
        # plt.scatter(*X_embedded)
        pal = sns.color_palette("hls", len(set(self.result)))
        # ann = [x for x,y in enumerate(X)]
        for i, txt in enumerate(self.indices):
            plt.scatter(X_embedded.T[0][i], X_embedded.T[1][i], color=pal[self.result[i]])
            plt.annotate(txt, (X_embedded.T[0][i], X_embedded.T[1][i]), color=pal[self.result[i]], alpha=0.2)
        plt.show()

    # plt.savefig('clus.png')

    def visualizeClustersAsHeatMaps(self) -> None:
        """
            This method visualizes the clusters as heatmaps.

            :return: void
        """
        # Show clusters as heatmaps (does not work too great for hand-written data)
        clusterdata = [[] for x in range(self.numClasses)]
        for idx, clus in enumerate(self.result):
            clusterdata[clus].append(idx)

        for cnum in range(len(clusterdata)):
            values = [self.distribution[idx] for idx in clusterdata[cnum]]
            fig = plt.figure(figsize=(10, 5))
            df = pd.DataFrame(values, index=clusterdata[cnum])
            plt.title('ClusterStore: ' + str(cnum))
            ax = sns.heatmap(df, center=0.0, xticklabels=False)
            ax.set_yticks(np.arange(len(clusterdata[cnum])))
            ax.set_yticklabels(clusterdata[cnum])
            plt.setp(ax.get_yticklabels(), rotation=0)
            plt.xlabel('Time ->')
            plt.ylabel('Trajectory id')

            plt.show()

    def simulateClusteringProcess(self) -> None:
        """
            This method makes multiple plots that replay the clustering process step-by-step.

            :return: void
        """
        # Simulates how the clustering happened
        # TODO: Fix the extra plots showing up at the end
        X_embedded_ = TSNE(random_state=42, n_components=2).fit_transform(self.distribution)

        for end in range(1, len(self.result)):

            fig = plt.figure(figsize=(18, 10))

            X_embedded = X_embedded_[0:end]
            ann = [x for x, y in enumerate(self.data)][0:end]

            pal = sns.color_palette("hls", len(set(self.result)))

            plt.subplot(1, 2, 1)
            sns.heatmap(self.distribution[0:end], center=0.0)

            plt.subplot(1, 2, 2)
            plt.scatter(X_embedded.T[0], X_embedded.T[1], color=[pal[c] for c in self.result[0:end]])
            for i, txt in enumerate(ann):
                plt.scatter(X_embedded.T[0][i], X_embedded.T[1][i], color=pal[self.result[i]])
                plt.annotate(txt, (X_embedded.T[0][i], X_embedded.T[1][i]), color=pal[self.result[i]])

            display.clear_output(wait=True)
            display.display(plt.gcf())
            time.sleep(0.01)  # change the rate of rendering



"""
	This module contains the interface for the data sources that are used by the 'SeqClu' algorithm to request
	data at every tick.
"""
from abc import ABC, abstractmethod
from typing import List, Union

from numpy import ndarray


class IDataSource(ABC):

	# PROPERTIES #

	@property
	@abstractmethod
	def classes(self) -> List[Union[chr, int]]:
		"""
			This property stores all the classes that are present in the data source.

			:return: All the classes that are present in the data set.
		"""
		pass

	@abstractmethod
	def advanceTick(self) -> List[Union[ndarray, list]]:
		"""
			This method advances the state of the data source by one tick. The method returns a list of sequences
			that should be processed by the 'SeqClu' algorithm during the next tick.

			:return: A list of sequences that should be processed by the 'SeqClu' algorithm during the next tick.
		"""
		pass



"""
    This package contains all interfaces that are used in the 'SeqClu' framework.
"""

from .baseline_prototypes import IBaselinePrototypes
from .candidate_store import ICandidateStore
from .data_generator import IDataGenerator
from .data_source import IDataSource
from .distance_measure import IDistanceMeasure
from .evaluator import IEvaluator
from .fake_data_source import IFakeDataSource
from .heuristics import IPrototypeValue
from .seqclu import ISeqClu

__all__ = ["IBaselinePrototypes",
           "ICandidateStore",
           "IDataGenerator",
           "IDataSource",
           "IDistanceMeasure",
           "IEvaluator",
           "IFakeDataSource",
           "IPrototypeValue",
           "ISeqClu"]



"""
	This module contains a method that constructs a data stream of sequenceStore
	from a given set of clusters, classes and number of prototypes.
"""
import random
from typing import Tuple, List, Union

from numpy import ndarray

Stream = List[Tuple[ndarray, Union[int, chr]]]


def constructStream(clusters: List[List[Union[ndarray, list]]], classes: List[Union[int, chr]], numPrototypes: int) \
		-> Tuple[Stream, Stream]:
	"""
		This method constructs a data stream of sequenceStore from a
		given set of clusters, classes and number of prototypes.

		:param clusters: The set of clusters that needs to be clustered by some clustering algorithm.
		:param classes: The identifier of the classes that belong to the clusters that were passed to this method.
		:param numPrototypes: The number of prototypes that will be used in the clustering algorithm.
		:return: A two-tuple where the first item is the complete stream and the second item is a stream of sequences
		from any of the clusters that are not prototypes where the order of sequenceStore is randomized.
	"""
	# Preparing input sequence data (Exp with different settings, e.g. randomize, inverted, etc)
	stream = []

	print(f"[SeqCluCLI] The number of sequences per class is {map(len, clusters)}.")

	# Q: How to select initial prototypes?
	for i, cluster in enumerate(clusters):
		stream.extend(list(zip(cluster[:numPrototypes], [classes[i]] * len(cluster))))

	print(f"[SeqCluCLI] The total number of prototypes is {len(stream)}.")

	# randomize incoming sequenceStore so they belong to random classes
	randomizedStream = []
	for i, cluster in enumerate(clusters):
		randomizedStream.extend(list(zip(cluster[numPrototypes:], [classes[i]] * len(cluster))))

	random.shuffle(randomizedStream)

	stream.extend(randomizedStream)

	return stream, randomizedStream



"""
    This module contains a command-line interface that can be used to run experiments on the HPC cluster.
"""
import json
import sys

from seqclupv.SeqClu import SeqClu
from seqclupv.SeqCluBaselineOnline import SeqCluBaselineOnline
from seqclupv.SeqCluBaselineOffline import SeqCluBaselineOffline
from seqclupv.library.data_generator.sine_curve_generator import CurveGenerator
from seqclupv.library.data_generator.handwritten_character_generator import HandwrittenCharacterGenerator
from seqclupv.library.data_generator.pebble_generator import GesturePebbleGenerator
from seqclupv.library.data_generator.plaid_generator import PLAIDGenerator
from seqclupv.library.data_source import FakeDataSource
from seqclupv.library.distance.dynamic_time_warping import DynamicTimeWarping
from seqclupv.library.evaluation.basic_evaluator import BasicEvaluator
from seqclupv.library.evaluation.basic_baseline_prototypes import BasicBaselinePrototypes
from seqclupv.library.heuristics.linear_prototype_value import LinearPrototypeValue


def areCurveParameters(parameters: list) -> bool:
    """
        This method verifies whether or not a list of parameters has the format of parameters that are used to
        generate a data set of sequences from a sine curve.

        :param parameters: A list of parameters for which needs to be verified whether or not it has the format of
        parameters that are used to generate a data set of sequences from a sine curve.
        :return: A boolean value indicating whether or not the given list of parameters has the format of parameters
        that are used to generate a data set of sequences from a sine curve.
    """
    if len(parameters) != 6:
        return False
    return type(parameters[0]) == int and type(parameters[1]) == float and type(parameters[2]) == float \
           and type(parameters[3]) == int and type(parameters[4]) == float and type(parameters[5]) == int


def areStringParameters(parameters: list) -> bool:
    """
        This method verifies whether or not all values in the list are strings.

        :param parameters: A list of parameters for which needs to be verified whether or not all values in it
        are strings.
        :return: A boolean value indicating whether or not all values in the given list of parameters are strings.
    """
    for parameter in parameters:
        if type(parameter) != str:
            return False
    return True


def areSeqCluParameters(parameters: list) -> bool:
    """
        This method verifies whether or not a list of parameters has the format of parameters that are used to
        configure the 'SeqClu-PV' algorithm.

        :param parameters: A list of parameters for which needs to be verified whether or not it has the format of
        parameters that are used to configure the 'SeqClu-PV' algorithm.
        :return: A boolean value indicating whether or not the given list of parameters has the format of parameters
        that are used to configure the 'SeqClu-PV' algorithm.
    """
    if len(parameters) != 5:
        return False
    return type(parameters[0]) == int and type(parameters[1]) == float \
           and type(parameters[2]) == float and type(parameters[3]) == bool and type(parameters[4]) == bool


def areTimeSeriesClassificationParameters(parameters: list) -> bool:
    """
        This method verifies whether or not a list of parameters has the format of parameters that are used to
        load a data set of sequences from data files downloaded from TimeSeriesClassification.com.

        :param parameters: A list of parameters for which needs to be verified whether or not it has the format of
        parameters that are used to load a data set of sequences from data files downloaded from
        TimeSeriesClassification.com.
        :return: A boolean value indicating whether or not the given list of parameters has the format of parameters
        that are used to load a data set of sequences from data files downloaded from TimeSeriesClassification.com.
    """
    if len(parameters) != 2:
        return False
    return type(parameters[0]) == bool and type(parameters[1]) == str


def parseBool(string: str) -> bool:
    """
        This method parses a string to a boolean.

        :param string: The string that needs to be parsed to a boolean.
        :return: The boolean equivalent of the passed string.
    """
    if string == "False":
        return False
    if string == "True":
        return True
    raise ValueError("Invalid argument passed to boolean constructor.")


def main(argv) -> None:
    """
        This method is the main method for the command-line interface of 'SeqClu-PV'. All arguments for the algorithm
        are passed via 'argv' and the algorithm is then configured and run.

        :param argv: The parameters that are required to configure the algorithm.
        :return: void
    """
    if len(argv) != 10:
        print("[SeqCluCLI] Too few or too many arguments. Shutting down.")
        return
    print(f"[SeqCluCLI] Passed arguments are {argv[1:]}.")
    numPrototypes = int(argv[1])
    numRepresentativePrototypes = int(argv[2])
    maxPerTick = int(argv[3])
    dataSourceParameters = json.loads(argv[4])
    seqCluParameters = json.loads(argv[5])
    maxIter = int(argv[6])
    online = parseBool(argv[7])
    onlySeqClu = parseBool(argv[8])
    experimentName = argv[9]

    numClusters = len(dataSourceParameters)

    if type(dataSourceParameters) != list:
        raise ValueError("Data source parameters must be stored in a list.")
    if len(dataSourceParameters) == 0:
        raise ValueError("Data source parameters cannot be empty.")

    if type(seqCluParameters) != list:
        raise ValueError("SeqClu parameters must be stored in a list.")
    if len(seqCluParameters) == 0:
        seqClu = False
    elif areSeqCluParameters(seqCluParameters):
        seqClu = True
    else:
        raise ValueError("SeqClu parameters are not in the right format.")

    # 'areStringParameters' can be interpreted as 'areHandwrittenParameters'.
    if areStringParameters(dataSourceParameters):
        classes = dataSourceParameters
        dataGenerator = HandwrittenCharacterGenerator(classes, numPrototypes)
        fakeDataSource = FakeDataSource(maxPerTick, [dataGenerator], numPrototypes, classes)
    elif areTimeSeriesClassificationParameters(dataSourceParameters):
        computeDistribution = dataSourceParameters[0]
        dataSetName = dataSourceParameters[1]
        if dataSetName == "pebble":
            dataGenerator = GesturePebbleGenerator(numPrototypes, computeDistribution)
        elif dataSetName == "plaid":
            dataGenerator = PLAIDGenerator(numPrototypes, computeDistribution)
        else:
            raise ValueError("Invalid data set name provided.")
        dataGenerator.generateData()
        numClusters = len(dataGenerator.classes)
        fakeDataSource = FakeDataSource(maxPerTick, [dataGenerator], numPrototypes, dataGenerator.classes)
    elif all(areCurveParameters(elem) for elem in dataSourceParameters):
        dataGenerators = []
        for parameters in dataSourceParameters:
            dataGenerators.append(CurveGenerator(parameters[0], (parameters[1], parameters[2]),
                                                 parameters[3], parameters[4], parameters[5]))
        classes = list(range(numClusters))
        fakeDataSource = FakeDataSource(maxPerTick, dataGenerators, numPrototypes, classes)
    else:
        raise ValueError("Data source parameters are not handwritten character or curve generator parameters.")

    if fakeDataSource.data is None:
        fakeDataSource.generateData()
    labels = fakeDataSource.labels
    print("[SeqCluCLI] Labels provided by data set are as follows.")
    print(labels)
    print("[SeqCluCLI] Labels including sequence hashes provided by fake data source are as follows.")
    print(fakeDataSource.actualLabels)

    distanceMeasure = DynamicTimeWarping()

    if seqClu:
        bufferSize = int(seqCluParameters[0])
        minimumRepresentativeness = float(seqCluParameters[1])
        prototypeScale = float(seqCluParameters[2])
        clusterAssignment = bool(seqCluParameters[3])
        buffering = bool(seqCluParameters[4])

        prototypeValue = LinearPrototypeValue(prototypeScale)
        seqClu = SeqClu(fakeDataSource, distanceMeasure, numClusters, numRepresentativePrototypes, numPrototypes,
                        bufferSize, minimumRepresentativeness, prototypeValue,
                        clusterAssignment, buffering)
    else:
        seqClu = None

    if not onlySeqClu:
        if online:
            seqCluBaselineOnline = SeqCluBaselineOnline(fakeDataSource, distanceMeasure, numClusters, numPrototypes)
            seqCluBaselineOffline = None
        else:
            seqCluBaselineOnline = None
            seqCluBaselineOffline = SeqCluBaselineOffline(fakeDataSource, distanceMeasure,
                                                          numClusters, numPrototypes, maxIter)
    else:
        seqCluBaselineOnline = None
        seqCluBaselineOffline = None
    BasicEvaluator(fakeDataSource, distanceMeasure, seqClu, seqCluBaselineOnline,
                   seqCluBaselineOffline, BasicBaselinePrototypes.getPrototypes(experimentName)).evaluate()


if __name__ == "__main__":
    main(sys.argv)



"""
    This package contains implementations of heuristic functions that are used in variants of the 'SeqClu' algorithm.
"""

from .linear_prototype_value import LinearPrototypeValue


__all__ = ["LinearPrototypeValue"]



"""
    This module contains a class that is responsible for generating gesture pebble data that are passed to the
    'SeqClu' algorithm.

    Link to the 'GesturePebbleZ1' data set
    http://www.timeseriesclassification.com/description.php?Dataset=GesturePebbleZ1
"""

import os
from typing import List, Union

from numpy import ndarray

from seqclupv.library.data_generator.time_series_classification_base import TimeSeriesClassificationBase


class GesturePebbleGenerator(TimeSeriesClassificationBase):

    # CONSTRUCTOR #

    def __init__(self, numPrototypes: int, computeDistances: bool) -> None:
        """
            This method initializes the data generator given the number of prototypes and a boolean value indicating
            whether or not the pair-wise distances between all items in the data set need to be computed. This
            data generator is applicable to the 'GesturePebbleZ1' data set.

            :param numPrototypes: The number of prototypes that will be used in the 'SeqClu' algorithm.
            :param computeDistances: A boolean value indicating whether or not the pair-wise distances between all
            items in the data set.
        """
        super().__init__(numPrototypes, computeDistances)

    # PUBLIC METHODS #

    def generateData(self) -> List[Union[ndarray, list]]:
        """
            This method generates a data set that is used in the 'SeqClu' algorithm or one of its extensions as
            a list of sequences.

            :return: A data set that is used in the 'SeqClu' algorithm or one of its extensions as
            a list of sequences.
        """
        path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "../../data/GesturePebbleZ1")
        data, labels = GesturePebbleGenerator.loadData(path, ["GesturePebbleZ1_TRAIN.ts", "GesturePebbleZ1_TEST.ts"])
        resultData, resultLabels = self._generateData(data, labels, ['1', '2', '3', '4', '5', '6'])
        return resultData



"""
    This module contains classes that bundle several approaches to visualize the results of the variations of
    the 'SeqClu' algorithm that are contained in the package.
"""
from .visualizer import Visualizer

__all__ = ["Visualizer"]



"""
    This module contains a method that calculates the F1-score given sets of true labels and predicted labels.
"""

from typing import Dict, Union

import numpy as np
from sklearn.metrics import f1_score


def calculate_f1_score(trueLabels: Dict[str, Union[chr, int]],
                       predictedLabels: Dict[str, Union[chr, int]], average: str) -> float:
    """
        This method calculates the F1-score given a set of true labels and predicted labels as well as the averaging
        procedure that decides the type of F1-score.

        :param trueLabels: The correct labels of the data as a dictionary where the keys are the hashes of the sequences
        and the values are the labels.
        :param predictedLabels: The predicted labels of the data as a dictionary where the keys are the hashes of the
        sequences and the values are the labels.
        :param average: The averaging procedure used in the statistical test. The averaging procedure decides the type
        of F1-score, this can be the micro, macro or weighted F1-score.
        :return: The F1-score for the given sets of true labels and predicted labels.
    """
    length = len(trueLabels)
    assert length == len(predictedLabels)

    trueLabelsArray = np.empty((length,), dtype=object)
    predictedLabelsArray = np.empty((length,), dtype=object)

    for i, key in enumerate(trueLabels):
        trueLabelsArray[i] = trueLabels[key]
        predictedLabelsArray[i] = predictedLabels[key]

    return f1_score(trueLabelsArray, predictedLabelsArray, average=average)



"""
    This module contains a base class that is responsible for generating time series data from
    https://www.timeseriesclassification.com that are passed to the 'SeqClu' algorithm.
"""

from abc import ABC
from typing import List, Dict, Tuple, Union
import os

from fastdtw import fastdtw
import numpy as np
from numpy import ndarray
from scipy.spatial.distance import euclidean
from sktime.utils.data_io import load_from_tsfile_to_dataframe

from seqclupv.library.interfaces.data_generator import IDataGenerator
from seqclupv.library.utilities.construct_stream import constructStream
from seqclupv.library.utilities.hash_sequence import hashSequence


class TimeSeriesClassificationBase(IDataGenerator, ABC):

    # STATIC METHODS #

    @staticmethod
    def loadData(path: str, filenames: List[str]) -> Tuple[List[ndarray], List[str]]:
        """
            This method loads the data from a given path to some directory and file name.

            :param path: The path to some directory, relative to the current working directory.
            :param filenames: The names of the files that should be loaded.
            :return: The data that was loaded from all of the files.
        """
        dataPath = os.path.join(os.getcwd(), path)

        data: List[ndarray] = []
        labels = np.array([])

        for filename in filenames:
            newData, newLabels = load_from_tsfile_to_dataframe(os.path.join(dataPath, filename))
            newData = newData["dim_0"]
            for x in newData:
                xArray = x.to_numpy()
                data.append(xArray)
            labels = np.concatenate((labels, newLabels))

        return data, labels

    # PROPERTIES #

    @property
    def computeDistances(self) -> bool:
        """
            This property stores a boolean value indicating whether or not the pair-wise distances between items in the
            data set should be computed.

            :return: A boolean value indicating whether or not the pair-wise distances between items in the
            data set should be computed.
        """
        return self._computeDistances

    # CONSTRUCTOR

    def __init__(self, numPrototypes: int, computeDistances: bool) -> None:
        """
            This method initializes the data generator given the number of prototypes that are drawn from the data set
            and a boolean value indicating whether or not the pair-wise distances between items in the data set should
            be computed.

            :param numPrototypes: The number of prototypes that should be drawn from the data set.
            :param computeDistances: A boolean value indicating whether or not the pair-wise distances between items in
            the data set should be computed.
        """
        super().__init__(numPrototypes)
        self._computeDistances = computeDistances

    # PRIVATE METHODS #

    def _generateData(self, data: List[ndarray], labels: List[str],
                      classes: List[str]) -> Tuple[List[Union[ndarray, list]], Dict[str, str]]:
        """
            This method creates the required data structures from the loaded data, labels and classes.

            :param data: The data that were loaded from one of the data files.
            :param labels: The labels of the data that were loaded from one of the data files.
            :param classes: All the classes that are present in the data set.
            :return: The constructed data stream and the labels of the data ordered according to
            the ordering of the items in the data stream.
        """
        labelResult = {}
        for i, label in enumerate(labels):
            if label in classes:
                labelResult[hashSequence(data[i])] = label

        clustersDict: Dict[str, List[ndarray]] = {}
        for clusterIdentifier in classes:
            clustersDict[clusterIdentifier] = []

        for i, sequence in enumerate(data):
            label = labels[i]
            if label in classes:
                clustersDict[label].append(sequence)

        clusters = []
        classDictionary: Dict[str, int] = {}
        for i, clusterIdentifier in enumerate(classes):
            clusters.append(clustersDict[clusterIdentifier])
            classDictionary[clusterIdentifier] = i

        self._classes = classes
        self._classDictionary = classDictionary

        trajectory, randomList = constructStream(clusters, self.classes, self.numPrototypes)

        self._data = [x for (x, y) in trajectory]  # Data: This is a list of sequences.
        self._indices = [x for x, y in enumerate(self.data)]  # IDs: This is a list of indices of the samples in X.
        self._labels = [y for (x, y) in trajectory]  # classes: This is a list of labels assigned to the samples in X.

        if self.computeDistances:
            distances = [-1] * len(self.data)
            distances = [[-1] * len(self.data) for _ in distances]

            for i in range(len(self.data)):
                for j in range(i + 1):
                    _d = None
                    if i == j:
                        distances[i][j] = 0.0
                    elif i > j:
                        _d = fastdtw(self.data[i], self.data[j], dist=euclidean)[0]
                        distances[i][j] = _d
                        distances[j][i] = _d

            self._distances = distances

        return self.data, labelResult



"""
    This module contains an implementation of a fake data source that feeds a random number of elements of a predefined
    data set to the 'SeqClu' algorithm at every tick.
"""

from random import randint
from typing import Dict, List, Optional, Union

from numpy import ndarray

from seqclupv.library.interfaces.data_generator import IDataGenerator
from seqclupv.library.interfaces.fake_data_source import IFakeDataSource
from seqclupv.library.utilities.construct_stream import constructStream
from seqclupv.library.utilities.hash_sequence import hashSequence


class FakeDataSource(IFakeDataSource, IDataGenerator):

    # STATIC METHODS #

    @staticmethod
    def combineGeneratedData(clusters: List[List[Union[ndarray, list]]], classes: List[Union[int, chr]],
                             numPrototypes: int) -> tuple:
        """
            This method combines the data that were generated from multiple data sources into one data stream.

            :param clusters: The logical clusters with all the different classes for all the data generators.
            :param classes: All the classes that should be present in the data provided by the data source.
            :param numPrototypes: The number of prototypes used in the 'SeqClu' algorithm.
            :return: The class dictionary, indices, labels and pair-wise distances of the data,
            as well as the data itself
        """
        classDictionary = {k: v for v, k in enumerate(classes)}
        trajectory, randomList = constructStream(clusters, classes, numPrototypes)

        print(f"[SeqCluCLI] The number of incoming sequences is {len(randomList)}.")
        print(f"[SeqCluCLI] The total number of sequences is {len(trajectory)}.")

        data = [x for (x, y) in trajectory]
        indices = [x for x, y in enumerate(data)]
        labels = [y for (x, y) in trajectory]

        return classDictionary, indices, labels, None, data

    # PROPERTIES #

    @property
    def actualLabels(self) -> Optional[Dict[str, Union[chr, int]]]:
        """
            This property stores the correct labels of the data in the data source.

            :return: The correct labels of the data in the data set.
        """
        return self._actualLabels

    @property
    def classes(self) -> List[Union[chr, int]]:
        """
            This property stores all the classes that are present in the data source.

            :return: All the classes that are present in the data set.
        """
        return self._classes

    @property
    def currentIndex(self) -> int:
        """
            This property stores the current index of the data provided by the data source.

            :return: An integer representing the current index of the data provided by the data source.
        """
        return self._currentIndex

    @property
    def currentTick(self) -> int:
        """
            This property stores the current tick of the algorithm.

            :return: An integer representing the current tick of the algorithm.
        """
        return self._currentTick

    @property
    def data(self) -> List[Union[ndarray, list]]:
        """
            This property stores the data provided by the data source.

            :return: The data provided by the data source.
        """
        return self._data

    @property
    def dataSize(self) -> Optional[int]:
        """
            This property stores the amount of data that are provided by the data source.

            :return: An integer representing the amount of data that are provided by the data source.
        """
        return self._dataSize

    @property
    def dataGenerators(self) -> List[IDataGenerator]:
        """
            This property stores the data generators that are used to generate the data provided by the data source.

            :return: A list of data generators that are used to generate the data provided by the data source.
        """
        return self._dataGenerators

    @property
    def maxPerTick(self) -> int:
        """
            This property stores the maximum number of data that can be provided by the data source per tick.

            :return: An integer representing the maximum number of data
            that can be provided by the data source per tick.
        """
        return self._maxPerTick

    @property
    def numClasses(self) -> int:
        """
            The number of classes that are present in the data provided by the data source.

            :return: An integer representing the number of classes that are
            present in the data provided by the data source.
        """
        return self._numClasses

    @property
    def numPrototypes(self) -> int:
        """
            The number of prototypes that are used in the 'SeqClu' algorithm.

            :return: An integer representing the number of prototypes that are used in the 'SeqClu' algorithm.
        """
        return self._numPrototypes

    @currentIndex.setter
    def currentIndex(self, value: int) -> None:
        """
            This is the setter for the 'currentIndex' property.

            :param value: The value that the 'currentIndex' property needs to be set to.
            :return: void
        """
        self._currentIndex = value

    @data.setter
    def data(self, value: ndarray) -> None:
        """
            This is the setter for the 'data' property.

            :param value: The value that the 'data' property needs to be set to.
            :return: void
        """
        self._data = value

    # CONSTRUCTOR #

    def __init__(self, maxPerTick: int, dataGenerators: List[IDataGenerator],
                 numPrototypes: int, classes: List[Union[chr, int]]) -> None:
        """
            This method initializes the data source given the maximum number of data per tick, the data generators
            that should be used to generate the data provided by the data source and the number of prototypes
            that is used in the 'SeqClu' algorithm.

            :param maxPerTick: The maximum number of data that can be provided by the data source per tick.
            :param dataGenerators: The data generators that should be used to
            generate the data provided by the data source.
            :param numPrototypes: The number of prototypes that is used in the 'SeqClu' algorithm.
            :param classes: All the classes that should be present in the data provided by the data source.
        """
        super().__init__(numPrototypes)
        self._actualLabels = None
        self._currentIndex = 0
        self._currentTick = -1
        self._dataGenerators = dataGenerators
        self._dataSize = None
        self._maxPerTick = maxPerTick
        self._classes = classes
        self._numClasses = len(self.classes)
        self._numPrototypes = numPrototypes

    # PUBLIC METHODS #

    def advanceTick(self) -> List[Union[ndarray, list]]:
        """
            This method advances the state of the data source by one tick. The method returns a list of sequences
            that should be processed by the 'SeqClu' algorithm during the next tick.

            :return: A list of sequences that should be processed by the 'SeqClu' algorithm during the next tick.
        """
        assert self.dataSize is not None
        print(f"[SeqClu] Current index of data is {self.currentIndex}.")
        numberOfElementsToFeed: int = randint(1, self.maxPerTick)
        print(f"[SeqClu] The number of elements that will be processed in this iteration is {numberOfElementsToFeed}.")
        toReturn = None
        if self.currentIndex + numberOfElementsToFeed >= self.dataSize:
            if self.currentIndex >= self.dataSize:
                numberOfElementsToFeed = 0
                toReturn = []
            else:
                numberOfElementsToFeed = self.dataSize - self.currentIndex
        if toReturn is None:
            toReturn = self.data[self.currentIndex:self.currentIndex + numberOfElementsToFeed]
        self._currentIndex += numberOfElementsToFeed
        return toReturn

    def generateData(self) -> List[Union[ndarray, list]]:
        """
            This method generates a data set that is used in the 'SeqClu' algorithm or one of its extensions as
            a list of sequences.

            :return: A data set that is used in the 'SeqClu' algorithm or one of its extensions as
            a list of sequences.
        """
        if len(self.dataGenerators) == 1:
            dataGenerator = self.dataGenerators[0]
            if dataGenerator.data is None:
                dataGenerator.generateData()
            self._data = dataGenerator.data
            self._classDictionary = dataGenerator.classDictionary
            self._classes = dataGenerator.classes
            self._distances = dataGenerator.distances
            self._indices = dataGenerator.indices
            self._labels = dataGenerator.labels
        else:
            clusters: List[List[Union[ndarray, list]]] = list(
                map(lambda x: x.generateData() if x.data is None else x.data, self.dataGenerators))
            self._classDictionary, self._indices, self._labels,\
                self._distances, self._data = FakeDataSource.combineGeneratedData(clusters, self.classes,
                                                                                  self.numPrototypes)
        self._dataSize = len(self.data)
        actualLabels = {}
        for i, label in enumerate(self.labels):
            actualLabels[hashSequence(self.data[i])] = label
        self._actualLabels = actualLabels
        return self.data

    def reset(self) -> None:
        """
            This method sets the current index of the data provided by the data source back to zero to reset the
            data source.

            :return: void
        """
        self._currentIndex = 0



"""
    This package contains classes that are responsible for storing information that is relevant to the clustering
    process that is part of the 'SeqClu' algorithm. The relevant information includes a set of candidate prototypes,
    a set of clusters and a set of prototypes for some cluster.
"""

from .candidates import CandidateStore
from .cluster import ClusterStore
from .prototype_frequencies import PrototypeFrequencyStore
from .prototypes import PrototypeStore
from .sequences import SequenceStore


__all__ = ["CandidateStore",
           "ClusterStore",
           "PrototypeFrequencyStore",
           "PrototypeStore",
           "SequenceStore"]



"""
    This module contains an implementation of the offline version of the 'SeqClu' algorithm that is based on
    Partitioning Around Medoids (PAM).
"""

import time
from typing import Dict, List, Tuple

import numpy as np
from numpy import ndarray

from seqclupv.library.data_source.fake_data_source import FakeDataSource
from seqclupv.library.interfaces.distance_measure import IDistanceMeasure
from seqclupv.library.interfaces.seqclu import ISeqClu
from seqclupv.library.utilities.hash_sequence import hashSequence


class SeqCluBaselineOffline(ISeqClu):

    # STATIC METHODS #

    @staticmethod
    def initializeClusters(numClusters: int, numPrototypes: int) -> ndarray:
        """
            This method initializes the clusters of the 'SeqClu' algorithm given a number of clusters and prototypes.

            :param numClusters: The number of clusters that should be used in the 'SeqClu' algorithm.
            :param numPrototypes: The number of prototypes that should be used in the 'SeqClu' algorithm.'
            :return: The indices of the prototypes for all the clusters, stored in a two-dimensional array.
        """
        result: ndarray = np.empty((numClusters, numPrototypes), dtype=int)
        for i in range(numClusters):
            result[i, :] = [j for j in range(i * numPrototypes, (i + 1) * numPrototypes)]
        return result

    # PROPERTIES #

    @property
    def currentConfigurationCost(self) -> float:
        """
            This property stores the cost of the current configuration.

            :return: A float representing the cost of the current configuration.
        """
        return self._currentConfigurationCost

    @property
    def data(self) -> List[ndarray]:
        """
            This property stores the data that are processed by the 'SeqClu' algorithm.

            :return: The data that are processed by the 'SeqClu' algorithm.
        """
        return self._data

    @property
    def dataSize(self) -> int:
        """
            This property stores the number of sequences that the data set contains.

            :return: An integer representing the number of sequences that the data set contains.
        """
        return self._dataSize

    @property
    def distances(self) -> ndarray:
        """
            This property stores the pair-wise distances for all pairs of sequences in the data set that is processed
            by the 'SeqClu' algorithm.

            :return: A two-dimensional array that contains the pair-wise distances for all pairs of sequences in the
            data set that is processed by the 'SeqClu' algorithm.
        """
        return self._distances

    @property
    def clusters(self) -> ndarray:
        """
            This property stores the current configuration as a two-dimensional array containing the indices of the
            prototypes for every cluster.

            :return: The current configuration as a two-dimensional array containing the indices of the
            prototypes for every cluster.
        """
        return self._clusters

    @property
    def finalLabels(self) -> Tuple[ndarray, Dict[str, int]]:
        """
            This property stores the correct labels as a dictionary where the keys are the hashes of the sequences
            and the values are the correct labels.

            :return: The correct labels as a dictionary where the keys are the hashes of the sequences
            and the values are the correct labels.
        """
        test = set([])
        resultTwo: Dict[str, int] = {}
        result = np.empty((self.dataSize,), dtype=int)
        for i, cluster in enumerate(self.clusters):
            for prototypeIdx in cluster:
                prototypeHash = hashSequence(self.data[prototypeIdx])
                resultTwo[prototypeHash] = self.classes[i]
                result[prototypeIdx] = i
                test.add(prototypeIdx)
        for i, nonPrototypeIdx in enumerate(self._getNonPrototypeIndices(self.clusters)):
            nonPrototypeHash = hashSequence(self.data[nonPrototypeIdx])
            resultTwo[nonPrototypeHash] = self.classes[self.labels[i]]
            result[nonPrototypeIdx] = self.labels[i]
            test.add(nonPrototypeIdx)
        return result, resultTwo

    @property
    def labels(self) -> ndarray:
        """
            This property stores the labels of all sequences that are not prototypes in the 'SeqClu' algorithm.

            :return: The labels of all sequences that are not prototypes in the 'SeqClu' algorithm.
        """
        return self._labels

    @property
    def maxTicks(self) -> int:
        """
            This property stores the maximum number of ticks the algorithm is allowed to execute.

            :return: An integer representing the maximum number of ticks the algorithm is allowed to execute.
        """
        return self._maxTicks

    @property
    def time(self) -> float:
        """
            This property stores the time it took the algorithm to initialize itself.

            :return: A float representing the number of seconds it took the algorithm to initialize itself.
        """
        return self._time

    # CONSTRUCTOR #

    def __init__(self, fakeDataSource: FakeDataSource, distanceMeasure: IDistanceMeasure,
                 numClusters: int, numPrototypes: int, maxTicks: int) -> None:
        """
            This method initializes the 'SeqClu' algorithm with a data source, distance measure, number of clusters
            and number of prototypes.

            :param fakeDataSource: The fake data source that should be used in the 'SeqClu' algorithm.
            :param distanceMeasure: The distance measure that should be used in the 'SeqClu' algorithm.
            :param numClusters: The number of clusters that should be used in the 'SeqClu' algorithm.
            :param numPrototypes: The number of prototypes that should be used in the 'SeqClu' algorithm.
            :param maxTicks: The maximum number of ticks that the algorithm is allowed to execute.
            :return: void
        """
        super().__init__(fakeDataSource, distanceMeasure, numClusters, numPrototypes)
        self._data = fakeDataSource.data
        self._dataSize = fakeDataSource.dataSize
        startTime = time.perf_counter()
        self._distances = np.empty((self.dataSize, self.dataSize))
        self._computeDistances()
        self._maxTicks = maxTicks
        self._clusters = SeqCluBaselineOffline.initializeClusters(numClusters, numPrototypes)
        self._labels, self._currentConfigurationCost = self._computeConfigurationCost(self.clusters,
                                                                                      self._getNonPrototypeIndices(self.clusters))
        endTime = time.perf_counter()
        self._time = endTime - startTime

    # PUBLIC METHODS #

    def execute(self) -> None:
        """
            This method executes the 'SeqClu' algorithm from start to finish. The main loop over the data is
            implemented in here.

            :return: void
        """
        for tick in range(self.maxTicks):
            print(f"[SeqCluBaselineOffline]: Tick {tick}")
            if self.finish:
                break

            oldClusters = self.clusters.copy()

            self._executeOptimalSwap(self._getNonPrototypeIndices(self.clusters))

            # Convergence is obtained once the prototypes for all clusters no longer change.
            if np.all(oldClusters == self.clusters):
                break

    # PUBLIC METHODS #

    def processSequence(self, sequence: Tuple[str, ndarray], considerCandidacy: bool) -> None:
        """
            Since this algorithm is the offline variant of the algorithm, sequences are not processed one-by-one, but
            all at once in an iterative fashion. This method should never be called.
        """
        raise ValueError("This function should never be called.")

    # PRIVATE METHODS #

    def _computeConfigurationCost(self, clusters: ndarray, nonPrototypeIndices: ndarray) -> Tuple[ndarray, float]:
        """
            This method computes the cost of a given configuration of cluster prototypes as well as the predicted
            labels for the given configuration.

            :param clusters: The configuration of cluster prototypes.
            :param nonPrototypeIndices: The indices of the sequences that are not prototypes.
            :return: The predicted labels for the given configuration and the cost for the given configuration.
        """
        resultCost = 0
        resultLabels = np.empty((len(nonPrototypeIndices),), dtype=int)
        for idx, nonPrototypeIdx in enumerate(nonPrototypeIndices):
            label, distance = self._lowestDistanceToCluster(clusters, nonPrototypeIdx)
            resultLabels[idx] = label
            resultCost += distance
        return resultLabels, resultCost

    def _computeDistances(self) -> None:
        """
            This method computes the pair-wise distances for all pairs of sequences in the data set.

            :return: void
        """
        length = len(self.data)
        for i, sequenceOne in enumerate(self.data):
            print(f"[SeqCluBaselineOffline] Computing distances is at iteration {i} of {length}.")
            for j, sequenceTwo in enumerate(self.data):
                if i == j:
                    self.distances[i][j] = 0
                    continue
                distance = self.distanceMeasure.calculateDistance(sequenceOne, sequenceTwo)
                self.distances[i][j] = distance
                self.distances[j][i] = distance

    def _distanceToCluster(self, prototypeIndices: ndarray, sequenceIdx: int) -> float:
        """
            This method computes the distance from some sequence, identified by its index, and a set of prototypes,
            identified by their indices.

            :param prototypeIndices: The indices of the prototypes for which the distance
            to the sequence needs to be computed.
            :param sequenceIdx: The index of the sequence for which the distance to the prototypes needs to be computed.
            :return: The distance from the given sequence to the given set of prototypes.
        """
        return np.sum(self.distances[sequenceIdx][prototypeIndices])

    def _executeOptimalSwap(self, nonPrototypeIndices: ndarray) -> None:
        """
            This method executes the optimal swap as defined in the Partitioning Around Medoids (PAM) algorithm.

            :param nonPrototypeIndices: The indices of the data that are not prototypes.
            :return: void
        """
        costImprovement = 0
        improvedConfiguration = None
        improvedLabels = None

        for prototypeIndices in self.clusters:
            for prototypeIdx in prototypeIndices:
                for nonPrototypeIdx in nonPrototypeIndices:
                    swapConfiguration = np.where(self.clusters == prototypeIdx, nonPrototypeIdx, self.clusters)
                    swapLabels, swapConfigurationCost = self._computeConfigurationCost(swapConfiguration,
                                                                                       self._getNonPrototypeIndices(swapConfiguration))
                    swapCostImprovement = self.currentConfigurationCost - swapConfigurationCost
                    if swapCostImprovement > 0 and swapCostImprovement > costImprovement:
                        costImprovement = swapCostImprovement
                        improvedConfiguration = swapConfiguration
                        improvedLabels = swapLabels

        if improvedConfiguration is not None and improvedLabels is not None:
            self._clusters = improvedConfiguration
            self._currentConfigurationCost -= costImprovement
            self._labels = improvedLabels

    def _getNonPrototypeIndices(self, clusters: ndarray) -> ndarray:
        """
            This method returns the indices of the data that are not prototypes.

            :param clusters: The clusters that are represented as 'numClusters' * 'numPrototypes' prototypes.
            :return: The indices of the data that are not prototypes stored in a NumPy array.
        """
        return np.delete(np.arange(self.dataSize), clusters.flatten())

    def _lowestDistanceToCluster(self, clusters: ndarray, sequenceIdx: int) -> Tuple[int, float]:
        """
            This method computes the lowest distance from a given sequence to any of the clusters.

            :param clusters: The clusters that should be considered.
            :param sequenceIdx: The index of the sequence.
            :return: A tuple containing the cluster index that minimizes the distance from the given sequence to the
            cluster as well as this distance.
        """
        lowestClusterIdx = -1
        lowestDistance = np.inf
        for cIdx in range(self.numClusters):
            distance = self._distanceToCluster(clusters[cIdx], sequenceIdx)
            if distance < lowestDistance:
                lowestClusterIdx = cIdx
                lowestDistance = distance
        return lowestClusterIdx, lowestDistance



"""
	This module contains a class that is responsible for generating handwritten character data that are passed to the
	'SeqClu' algorithm.

	Link to the 'UJI Pen Character' data set
	https://archive.ics.uci.edu/ml/datasets/UJI+Pen+Characters

	NOTE: The characters that are contained in the data set are as follows.
	['C', 'U', 'V', 'W', 'S', 'O', '1', '2', '3', '5', '6', '8', '9']
"""

import glob
import os
import re
from typing import Dict, List, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from numpy import ndarray
from scipy.spatial.distance import euclidean
from fastdtw import fastdtw

from seqclupv.library.interfaces.data_generator import IDataGenerator
from seqclupv.library.utilities.construct_stream import constructStream


def parseFile(lines: List[str]) -> Dict[str, List[Tuple[int, int]]]:
	"""
		This method parses a file containing a data set of sequences. The lines in this file are processed in the
		method to obtain a dictionary in which the keys represent the name of some class and the values represent
		a list of sequences that belong to this class.

		:param lines: The lines that are present in some file containing a data set of sequences.
		:return: A dictionary in which the keys represent the name of some class and the values represent
		a list of sequences that belong to this class.
	"""
	points: Dict[str, List[Tuple[int, int]]] = dict()
	newChar = False
	cont = False
	point = []
	sequenceClass = None

	for line in lines[1:]:
		if '.COMMENT' in line and 'Class' in line and '[' in line and '#' not in line:
			b = re.findall('.*?\.COMMENT\s+Class\s+\[(.*?)\]', line)
			sequenceClass = b[0]
			newChar = True
			point = []
			continue
		if '.PEN_UP' in line:
			cont = False
			if sequenceClass not in points.keys():
				points[sequenceClass] = []
			points[sequenceClass].append(point)
		if '.PEN_DOWN' in line:
			cont = True
			continue
		if newChar and cont:
			b = re.findall('.*?(\d+)\s+([-\d]+).*', line)
			xy = b[0]
			point.append((int(xy[0]), int(xy[1])))

	return points


class HandwrittenCharacterGenerator(IDataGenerator):

	# CONSTRUCTOR #

	def __init__(self, classes: List[chr], numPrototypes: int) -> None:
		"""
			This method initializes the data generator with a given list of classes that the
			generated data set should contain.

			:param classes: The classes that are present in the data set as a list of characters.
			:param numPrototypes: The number of prototypes that will be used in the 'SeqClu' algorithm.
			:return: void
		"""
		super().__init__(numPrototypes)
		self._classes = classes

	# PUBLIC METHODS #

	def generateData(self) -> List[Union[ndarray, list]]:
		"""
			This method generates a data set that is used in the 'SeqClu' algorithm or one of its extensions as
			a list of sequences.

			:return: A data set that is used in the 'SeqClu' algorithm or one of its extensions as
			a list of sequences.
		"""
		self._classDictionary = {k: v for v, k in enumerate(self.classes)}

		segments = dict()
		files = glob.glob(os.path.join(os.path.dirname(os.path.abspath(__file__)), "../../data/handwriting/*"))
		for f in files:
			f_ = open(f, 'r')
			lines = f_.readlines()
			content = parseFile(lines)
			for classIdentifier, segment in content.items():
				if classIdentifier not in self.classes:
					continue
				if classIdentifier not in segments.keys():
					segments[classIdentifier] = []
				for sequence in segment:
					sequenceArray = np.array(sequence)
					identical = False
					for otherSequence in segments[classIdentifier]:
						if np.array_equal(sequenceArray, otherSequence):
							identical = True
							break
					if not identical:
						segments[classIdentifier].append(sequence)
			f_.close()

		clusters = []
		for classIdentifier in self.classes:
			clusters.append(segments[classIdentifier])

		# Visualize input characters
		for classIdentifier in segments.keys():
			print(f"[SeqCluCLI] Class \'{classIdentifier}\': {len(segments[classIdentifier])} instances.")
			fig, ax = plt.subplots()
			plt.title("Character: " + classIdentifier)

			for segment in segments[classIdentifier]:
				plt.plot([x[0] for x in segment], [x[1] for x in segment])
			ax.set_ylim(ax.get_ylim()[1], ax.get_ylim()[0])

		trajectory, randomList = constructStream(clusters, self.classes, self.numPrototypes)

		print(f"[SeqCluCLI] The number of incoming sequences is {len(randomList)}.")
		print(f"[SeqCluCLI] The total number of sequences is {len(trajectory)}.")

		self._data = [x for (x, y) in trajectory]  # Data: This is a list of sequences.
		self._transformDataToArray()
		self._indices = [x for x, y in enumerate(self.data)]  # IDs: This is a list of indices of the samples in X.
		self._labels = [y for (x, y) in trajectory]  # classes: This is a list of labels assigned to the samples in X.

		distances = [-1] * len(self.data)
		distances = [[-1] * len(self.data) for _ in distances]

		for i in range(len(self.data)):
			for j in range(i + 1):
				_d = None
				if i == j:
					distances[i][j] = 0.0
				elif i > j:
					_d = fastdtw(self.data[i], self.data[j], dist=euclidean)[0]
					distances[i][j] = _d
					distances[j][i] = _d

		self._distances = distances

		return self.data

	# PRIVATE METHODS #

	def _transformDataToArray(self) -> None:
		"""
			This method transforms all the data stored in the 'data' property to a list of NumPy arrays.

			:return: void
		"""
		result = []
		for sequence in self.data:
			arr = np.empty((len(sequence), 2))
			for i, (itemOne, itemTwo) in enumerate(sequence):
				arr[i][0] = itemOne
				arr[i][1] = itemTwo
			result.append(arr)
		self._data = result



"""
    This module contains a command-line interface for carrying out a statistical test on two provided samples of data.
    The function to carry out a statistical test on two provided samples is contained in this module as well.
"""

import json
import sys
from typing import List

from numpy import ndarray
from scipy.stats import wilcoxon, mannwhitneyu


def statistical_test(sampleOne: ndarray, sampleTwo: ndarray, testName: str) -> None:
    """
        This method carries out some statistical test, identified by its name, on two given samples.

        :param sampleOne: The first sample that should be used in the statistical test.
        :param sampleTwo: The second sample that should be used in the statistical test.
        :param testName: The name of the statistical test that should be used.
        :return: void
    """
    if testName == "wilcoxon":
        _, p = wilcoxon(sampleOne, sampleTwo)
    elif testName == "mannwhitneyu":
        _, p = mannwhitneyu(sampleOne, sampleTwo)
    else:
        raise ValueError("[StatTestCLI] Invalid test name provided.")

    if p > 0.05:
        print(f"[StatTestCLI] H0: There is no significant difference (p-value {p}).")
    else:
        print(f"[StatTestCLI] H1: There is a significant difference (p-value {p}).")


def main(argv: List[str]) -> None:
    """
        This method is the command-line interface for the statistical test program.

        :param argv: The arguments that are passed to the statistical test program.
        :return: void
    """
    if len(argv) != 4:
        print("[StatTestCLI] Too few or too many arguments. Shutting down.")
        return
    samplesOne = json.loads(argv[1])
    samplesTwo = json.loads(argv[2])
    test = argv[3]

    for sampleOne, sampleTwo in zip(samplesOne, samplesTwo):
        statistical_test(sampleOne, sampleTwo, test)


if __name__ == "__main__":
    main(sys.argv)



"""
    This module contains an interface for the 'SeqClu' algorithm. Some methods are already implemented.
"""

from abc import ABC, abstractmethod
from typing import Dict, List, Tuple, Union

from numpy import ndarray

from seqclupv.library.interfaces.data_source import IDataSource
from seqclupv.library.interfaces.distance_measure import IDistanceMeasure
from seqclupv.library.utilities.hash_sequence import hashSequence


class ISeqClu(ABC):

    # PROPERTIES #

    @property
    def classes(self) -> List[Union[chr, int]]:
        """
            This property stores the classes of the items that are present in the data set.

            :return: The classes of the items that are present in the data set.
        """
        return self._classes

    @property
    def dataSource(self) -> IDataSource:
        """
            This property stores the data source that is used to retrieve data that need to be processed by
            the algorithm.

            :return: The data source that is used to retrieve data that need to be processed by the algorithm.
        """
        return self._dataSource

    @property
    def distanceMeasure(self) -> IDistanceMeasure:
        """
            This property stores the distance measure that is used to compute the distance between two sequences.

            :return: The distance measure that is used to compute the distance between two sequences.
        """
        return self._distanceMeasure

    @property
    @abstractmethod
    def finalLabels(self) -> Dict[str, Union[chr, int]]:
        """
            This property stores the correct labels as a dictionary where the keys are the hashes of the sequences
            and the values are the correct labels.

            :return: The correct labels as a dictionary where the keys are the hashes of the sequences
            and the values are the correct labels.
        """
        pass

    @property
    def finish(self) -> bool:
        """
            This property stores a boolean value indicating whether or not the algorithm finished executing.

            :return: A boolean value indicating whether or not the algorithm finished executing.
        """
        return self._finish

    @property
    def numClusters(self) -> int:
        """
            This property stores the number of clusters that are used in the 'SeqClu' algorithm.

            :return: An integer representing the number of clusters that are used in the 'SeqClu' algorithm.
        """
        return self._numClusters

    @property
    def numPrototypes(self) -> int:
        """
            This property stores the number of prototypes that are used in the 'SeqClu' algorithm.

            :return: An integer representing the number of prototypes that are used in the 'SeqClu' algorithm.
        """
        return self._numPrototypes

    @property
    def tick(self) -> int:
        """
            This property stores the tick of the algorithm, which represents the moment in time relative to the start
            of the execution of the algorithm.

            :return: An integer representing the tick of the algorithm, which represents the moment in time relative
            to the start of the execution of the algorithm.
        """
        return self._tick

    @finish.setter
    def finish(self, value: bool) -> None:
        """
            This is the setter for the 'finish' property.

            :param value: The value that the 'finish' property should be set to.
            :return: void
        """
        self._finish = value

    # CONSTRUCTOR #

    def __init__(self, dataSource: IDataSource, distanceMeasure: IDistanceMeasure,
                 numClusters: int, numPrototypes: int) -> None:
        """
            This method initializes the 'SeqClu' algorithm with a data source, distance measure, number of clusters
            and number of prototypes.

            :param dataSource: The data source that should be used in the 'SeqClu' algorithm.
            :param distanceMeasure: The distance measure that should be used in the 'SeqClu' algorithm.
            :param numClusters: The number of clusters that should be used in the 'SeqClu' algorithm.
            :param numPrototypes: The number of prototypes that should be used in the 'SeqClu' algorithm.
            :return: void
        """
        self._classes = dataSource.classes
        self._dataSource = dataSource
        self._distanceMeasure = distanceMeasure
        self._finish = False
        self._numClusters = numClusters
        self._numPrototypes = numPrototypes
        self._tick = -1

    # PUBLIC METHODS #

    @abstractmethod
    def execute(self) -> None:
        """
            This method executes the 'SeqClu' algorithm from start to finish. The main loop over the data is
            implemented in here.

            :return: void
        """
        pass

    @abstractmethod
    def processSequence(self, sequence: Tuple[str, ndarray], considerCandidacy: bool) -> None:
        """
            This method processes one sequence of the data set that is processed in the 'SeqClu' algorithm.

            :param sequence: The sequence that should be processed in the 'SeqClu' algorithm.
            :param considerCandidacy: A boolean value indicating whether or not the algorithm should consider if the
            sequence could be a prototype for any of the clusters.
            :return: void
        """
        pass

    # PRIVATE METHODS #

    def _advanceTick(self) -> bool:
        """
            This method advances the tick of the algorithm.

            :return: A boolean value indicating whether or not sequences were processed during this tick.
        """
        # Increment the tick.
        self._tick += 1

        sequencesToProcess = self.dataSource.advanceTick()
        if not sequencesToProcess:
            return True

        for sequence in sequencesToProcess:
            sequenceHash = hashSequence(sequence)
            self.processSequence((sequenceHash, sequence), True)

        return False



"""
    This module is an implementation of an evaluator that measures the accuracy of the results obtained from the
    'SeqClu' algorithm as well as various other metrics such as the silhouette score, the number of distance
    computations that were made and the time taken to execute the algorithm.
"""
import time
from typing import Dict, List, Optional, Union

from seqclupv.SeqClu import SeqClu
from seqclupv.SeqCluBaselineOnline import SeqCluBaselineOnline
from seqclupv.SeqCluBaselineOffline import SeqCluBaselineOffline
from seqclupv.library.data_source import FakeDataSource
from seqclupv.library.distance.dynamic_time_warping import DynamicTimeWarping
from seqclupv.library.interfaces.evaluator import IEvaluator
from seqclupv.library.utilities.calculate_f1_score import calculate_f1_score
from seqclupv.library.utilities.hash_sequence import hashSequence


class BasicEvaluator(IEvaluator):

    # PROPERTIES #

    @property
    def actualLabels(self) -> Dict[str, Union[chr, int]]:
        """
            This property stores the correct labels of the data processed by the 'SeqClu' algorithm in a dictionary
            where the keys are the hashes of the data.

            :return: A dictionary where the keys are the hashes of the data and the values are the correct labels of the
            data processed by the 'SeqClu' algorithm.
        """
        return self._actualLabels

    @property
    def baselinePrototypes(self) -> Optional[List[List[str]]]:
        """
            This property stores the prototypes of the baseline that are used in the evaluator.

            :return: The prototypes of the baseline that are used in the evaluator.
        """
        return self._baselinePrototypes

    @property
    def fakeDataSource(self) -> FakeDataSource:
        """
            This property stores the fake data source that is used in the evaluator.

            :return: The fake data source that is used in the evaluator.
        """
        return self._fakeDataSource

    @property
    def distanceMeasure(self) -> DynamicTimeWarping:
        """
            This property stores the distance measure that is used in the 'SeqClu' algorithm.

            :return: The distance measure that is used in the 'SeqClu' algorithm.
        """
        return self._distanceMeasure

    @property
    def seqClu(self) -> Optional[SeqClu]:
        """
            This property stores the 'SeqClu' algorithm that is used in the evaluator.

            :return: The 'SeqClu' algorithm that is used in the evaluator.
        """
        return self._seqClu

    @property
    def seqCluBaselineOffline(self) -> Optional[SeqCluBaselineOffline]:
        """
            This property stores the offline baseline variant of the 'SeqClu' algorithm that is used in the evaluator.

            :return: The offline baseline variant of the 'SeqClu' algorithm that is used in the evaluator.
        """
        return self._seqCluBaselineOffline

    @property
    def seqCluBaselineOnline(self) -> Optional[SeqCluBaselineOnline]:
        """
            This property stores the online baseline variant of the 'SeqClu' algorithm that is used in the evaluator.

            :return: The online baseline variant of the 'SeqClu' algorithm that is used in the evaluator.
        """
        return self._seqCluBaselineOnline

    # CONSTRUCTOR #

    def __init__(self, fakeDataSource: FakeDataSource,
                 distanceMeasure: DynamicTimeWarping,
                 seqClu: Optional[SeqClu],
                 seqCluBaselineOnline: Optional[SeqCluBaselineOnline],
                 seqCluBaselineOffline: Optional[SeqCluBaselineOffline],
                 baselinePrototypes: Optional[List[List[str]]]) -> None:
        """
            This method initializes the evaluator.

            :param fakeDataSource: The fake data source that is used in the evaluator.
            :param distanceMeasure: The distance measure that is used in the 'SeqClu' algorithm.
            :param seqClu: The 'SeqClu' algorithm that is used in the evaluator.
            :param seqCluBaselineOnline: The online baseline variant of the 'SeqClu' algorithm that is used in the evaluator.
            :param seqCluBaselineOffline: The offline baseline variant of the 'SeqClu' algorithm that is used in the evaluator.
            :param baselinePrototypes: The prototypes of the baseline that are used in the evaluator.
            :return: void
        """
        self._actualLabels = fakeDataSource.actualLabels
        self._baselinePrototypes = baselinePrototypes
        self._fakeDataSource = fakeDataSource
        self._distanceMeasure = distanceMeasure
        self._seqClu = seqClu
        self._seqCluBaselineOffline = seqCluBaselineOffline
        self._seqCluBaselineOnline = seqCluBaselineOnline

    # PUBLIC METHODS #

    def evaluate(self) -> None:
        """
            This method evaluates the performance of the variants of the 'SeqClu' algorithm.

            :return: void
        """
        if self.seqCluBaselineOnline is not None:
            startTime = time.perf_counter()
            self.seqCluBaselineOnline.execute()
            endTime = time.perf_counter()
            timeTaken = endTime - startTime
            seqCluBaselineOnlineResult = self.seqCluBaselineOnline.finalLabels
            prototypeKeys = []
            for prototypeDict in self.seqCluBaselineOnline.prototypes:
                prototypeKeys.append(list(prototypeDict.keys()))
            self._printResults(seqCluBaselineOnlineResult, "SeqCluBaselineOnline", timeTaken, prototypeKeys)

            correct = 0
            if self.baselinePrototypes is not None:
                for i, baselineClusterPrototypes in enumerate(self.baselinePrototypes):
                    prototypes = self.seqCluBaselineOnline.clusters[i]
                    correct += len(set(prototypes).intersection(set(baselineClusterPrototypes)))
            prototypeAccuracy = correct / (self.seqClu.numPrototypes * self.seqClu.numClusters)
            print(f"[SeqClu] PrototypeAccuracy {prototypeAccuracy}")

        if self.seqCluBaselineOffline is not None:
            startTime = time.perf_counter()
            self.seqCluBaselineOffline.execute()
            endTime = time.perf_counter()
            timeTaken = endTime - startTime + self.seqCluBaselineOffline.time
            _, seqCluBaselineOfflineResult = self.seqCluBaselineOffline.finalLabels
            prototypeKeys = []
            for i, cluster in enumerate(self.seqCluBaselineOffline.clusters):
                prototypeKeysItem = []
                for prototypeIdx in cluster:
                    prototypeHash = hashSequence(self.seqCluBaselineOffline.data[prototypeIdx])
                    prototypeKeysItem.append(prototypeHash)
                prototypeKeys.append(prototypeKeysItem)
            self._printResults(seqCluBaselineOfflineResult, "SeqCluBaselineOffline", timeTaken, prototypeKeys)

        if self.seqClu is not None:
            startTime = time.perf_counter()
            self.seqClu.execute()
            endTime = time.perf_counter()
            timeTaken = endTime - startTime
            seqCluResult = self.seqClu.finalLabels
            prototypeKeys = []
            for cluster in self.seqClu.clusters:
                prototypeKeys.append(list(cluster.prototypes.prototypes.keys()))
            print(f"[SeqClu] Buffered sequences are {self.seqClu.bufferedSequences}")
            print(f"[SeqClu] Sequences clustered by approximation are {self.seqClu.clusteredByApproximation}")
            self._printResults(seqCluResult, "SeqClu", timeTaken, prototypeKeys)
            correctSimple = 0
            correctComplex = 0
            wrongComplex = 0
            total = len(self.seqClu.clusteredByApproximation)
            for sequenceHash in self.seqClu.clusteredByApproximation:
                actualLabel = self.actualLabels[sequenceHash]
                if self.seqClu.finalLabels[sequenceHash] == actualLabel:
                    correctSimple += 1
                    if self.seqCluBaselineOnline is not None and self.seqCluBaselineOnline.finalLabels[sequenceHash] != actualLabel:
                        correctComplex += 1
                elif self.seqCluBaselineOnline is not None and self.seqCluBaselineOnline.finalLabels[sequenceHash] == actualLabel:
                    wrongComplex += 1
            if total > 0:
                accuracySimple = correctSimple / total
                accuracyComplex = correctComplex / total
                accuracyWrongComplex = wrongComplex / total
                print(f"[SeqClu] Approximation: {accuracySimple},{accuracyComplex},{accuracyWrongComplex}")

            correct = 0
            if self.baselinePrototypes is not None:
                for i, baselineClusterPrototypes in enumerate(self.baselinePrototypes):
                    prototypes = self.seqClu.clusters[i].prototypes.prototypes
                    correct += len(set(prototypes.keys()).intersection(set(baselineClusterPrototypes)))
            prototypeAccuracy = correct / (self.seqClu.numPrototypes * self.seqClu.numClusters)
            print(f"[SeqClu] PrototypeAccuracy {prototypeAccuracy}")

            if self.seqCluBaselineOnline is not None:
                correctSimple = 0
                correctComplex = 0
                wrongComplex = 0
                total = len(self.seqClu.bufferedSequences)
                for sequenceHash in self.seqClu.bufferedSequences:
                    actualLabel = self.actualLabels[sequenceHash]
                    if self.seqClu.finalLabels[sequenceHash] == actualLabel:
                        correctSimple += 1
                        if self.seqCluBaselineOnline is not None and self.seqCluBaselineOnline.finalLabels[sequenceHash] != actualLabel:
                            correctComplex += 1
                    elif self.seqCluBaselineOnline is not None and self.seqCluBaselineOnline.finalLabels[sequenceHash] == actualLabel:
                        wrongComplex += 1
            else:
                correctSimple = 0
                correctComplex = 0
                wrongComplex = 0
            if total > 0:
                accuracySimple = correctSimple / total
                accuracyComplex = correctComplex / total
                accuracyWrongComplex = wrongComplex / total
                print(f"[SeqClu] Buffering: {accuracySimple},{accuracyComplex},{accuracyWrongComplex}.")
                print(f"[SeqClu] Buffering correctSimple, correctComplex, wrongComplex, "
                      f"total {correctSimple}, {correctComplex}, {wrongComplex}, {total}")

    # PRIVATE METHODS #

    def _printResults(self, result: Dict[str, Union[chr, int]], algorithmName: str,
                      timeTaken: float, prototypeHashes: List[List[str]]) -> None:
        """
            This method prints the results of one of the variants of the 'SeqClu' algorithm.

            :param result: The result obtained at the end of executing the variant of the 'SeqClu' algorithm.
            The result is a dictionary of the hashes of the sequences paired with the label of the sequence.
            :param algorithmName: The name of the variant of the 'SeqClu' algorithm.
            :param timeTaken: The time taken to execute the variant of the 'SeqClu' algorithm.
            :param prototypeHashes: The hashes of the prototypes obtained at the end
            of executing the variant of the 'SeqClu' algorithm.
            :return: void
        """
        print(f"[{algorithmName}] The results of the implementation of {algorithmName} are as follows.")
        print(f"[{algorithmName}] The result provided by the \'{algorithmName}\' algorithm is as follows.")
        print(result)
        print(f"[{algorithmName}] The amount of labels provided is {len(result)}.")
        seqCluResultSet = set(result.items())
        correct = len(seqCluResultSet.intersection(set(self.actualLabels.items())))
        print(f"[{algorithmName}] The amount of correct predictions is {correct}.")
        mistakes = self.fakeDataSource.dataSize - correct
        dataSize = self.fakeDataSource.dataSize
        accuracy = correct / self.fakeDataSource.dataSize
        timesCalled = self.distanceMeasure.timesCalled
        f1macro = calculate_f1_score(self.actualLabels, result, 'macro')
        f1micro = calculate_f1_score(self.actualLabels, result, 'micro')
        f1weighted = calculate_f1_score(self.actualLabels, result, 'weighted')
        print(f"[SeqClu] The algorithm made {mistakes} out "
              f"of {dataSize} mistakes.")
        print(f"[SeqClu] The amount of times the distance computation has been called is "
              f"{timesCalled}.")
        print(f"[SeqClu] the {algorithmName} algorithm took {timeTaken} seconds to complete.")
        print(f"[SeqClu] The clustering accuracy is therefore {accuracy}")
        print(f"[SeqClu] The F1 score (macro) is {f1macro}")
        print(f"[SeqClu] The F1 score (micro) is {f1micro}")
        print(f"[SeqClu] The F1 score (weighted) is {f1weighted}")
        print(f"[SeqClu] The prototype hashes are {prototypeHashes}")
        print(f"[SeqCluCSV] {timesCalled},{timeTaken},{accuracy},{f1macro},{f1micro},{f1weighted}")
        self.distanceMeasure.reset()
        self.fakeDataSource.reset()



"""
    This module contains a class called 'CandidateStore' that stores the candidate prototypes that are used in the
    'SeqClu' algorithm.
"""

from typing import Dict, Optional, Set, Tuple

from numpy import ndarray

from seqclupv.library.interfaces.candidate_store import ICandidateStore


class CandidateStore(ICandidateStore):

    # PROPERTIES #

    @property
    def candidates(self) -> Dict[str, Tuple[ndarray, Set[int]]]:
        """
            This method is a property that returns the stored set of candidateStore. The candidateStore are stored in a
            dictionary where key is the hash of the candidate and the value is a tuple containing the candidate itself
            and a set of cluster identifiers for which this sequence is a candidate prototype.

            :return: A dictionary where key is the hash of the candidate and the value is a tuple containing the
            candidate itself and a set of cluster identifiers for which this sequence is a candidate prototype.
        """
        return self._candidates

    @property
    def candidateHistory(self) -> Dict[str, int]:
        """
            This method is a property that returns a dictionary where the key is the hash of some candidate and the
            value is the tick at which the candidate was added to the buffer.

            :return: A dictionary where the key is the hash of some candidate and the
            value is the tick at which the candidate was added to the buffer.
        """
        return self._candidateHistory

    @property
    def lastUpdate(self) -> int:
        """
            This method is a property that returns the tick at which the buffer of candidateStore was last updated.

            :return: The tick at which the buffer of candidateStore was last updated.
        """
        return self._lastUpdate

    # CONSTRUCTOR #

    def __init__(self, numRepresentativePrototypes: int, numPrototypes: int, tick: int) -> None:
        """
            This method initializes the buffer of candidateStore with a given number of representative prototypes, total
            number of prototypes, the maximum size of the buffer of candidateStore and the tick at which the buffer is
            initialized.

            :param numRepresentativePrototypes: The number of representative prototypes.
            :param numPrototypes: The total number of prototypes.
            :param tick: The tick at which the buffer of candidateStore is initialized.
        """
        assert 0 < numRepresentativePrototypes < numPrototypes and numPrototypes > 0 and tick >= -1
        self._candidates = {}
        self._candidateHistory = {}
        self._lastUpdate = tick

    # PUBLIC METHODS #

    def addToCandidates(self, candidate: Tuple[Optional[str], Optional[ndarray]],
                        candidateFor: Set[int], tick: int) -> None:
        """
            This method adds an incoming sequence to the buffer of candidateStore.

            :param candidate: The incoming sequence that needs to be added to the buffer of candidateStore.
            :param candidateFor: The identifiers of the clusters that the incoming sequence is a candidate for.
            :param tick: The tick at which the incoming sequence is promoted to a candidate.
            :return: void
        """
        candidateHash, candidateData, = candidate
        assert candidateHash not in self.candidates

        self._candidates[candidateHash] = candidateData, candidateFor
        self._candidateHistory[candidateHash] = tick
        self._lastUpdate = tick

    def getCandidate(self, candidateHash: str) -> Tuple[ndarray, Set[int]]:
        """
            This method returns a candidate given its hash.

            :param candidateHash: The hash of the candidate that is requested.
            :return: A two-tuple containing the candidate itself and the identifiers of the clusters that the sequence
            is a candidate for.
        """
        # A requested candidate identified by its hash must always be available in the 'candidateStore' data structure.
        assert candidateHash in self.candidates
        return self.candidates[candidateHash]

    def lastUpdateCandidate(self, candidateHash: str) -> int:
        """
            This method returns the tick at which a candidate that is stored in the buffer was last updated.

            :param candidateHash: The hash of the candidate for which the information about when the candidate was last
            updated is requested.
            :return: The tick at which a candidate that is stored in the buffer was last updated.
        """
        # Since the data structure will only keep track of current candidateStore and this function is only intended
        # for use with candidateStore, the constraint that the provided hash of the candidateStore needs to be contained
        # in the data structure that stores the ticks at which candidateStore were updated is added.
        assert candidateHash in self.candidateHistory
        return self.candidateHistory[candidateHash]

    def removeFromCandidates(self, candidateHash: str) -> None:
        """
            This method removes a sequence from the set of candidateStore. This occurs when the sequence is fully processed
            and either promoted to a prototype or discarded.

            :param candidateHash: The hash of the candidate.
            :return: void
        """
        assert candidateHash in self.candidates

        del self.candidates[candidateHash]
        del self.candidateHistory[candidateHash]



"""
	This module contains an interface for the distance measures that are used to compute the distance between two
	data points in the 'SeqClu' algorithm.
"""
from abc import ABC, abstractmethod
from typing import Union

from numpy import ndarray


class IDistanceMeasure(ABC):

	@abstractmethod
	def calculateDistance(self, sequenceOne: Union[ndarray, list], sequenceTwo: Union[ndarray, list]) -> float:
		"""
			This method calculates the distance between two sequences.

			:param sequenceOne: The first sequence for which the distance should be computed.
			:param sequenceTwo: The second sequence for which the distance should be computed.
			:return: The distance between the two sequences.
		"""
		pass



"""
	This module contains an implementation of the 'Dynamic Time Warping' distance measure that is used in the 'SeqClu'
	algorithm to compute the distance between two sequences.
"""

from seqclupv.library.interfaces.distance_measure import IDistanceMeasure

from numpy import ndarray
from scipy.spatial.distance import euclidean
from fastdtw import fastdtw


class DynamicTimeWarping(IDistanceMeasure):

	# PROPERTIES #

	@property
	def timesCalled(self) -> int:
		"""
			This property stores the counter on the amount of times that the distance measure is used to compute the
			distance between two sequences.

			:return: An integer representing the counter on the amount of times that the distance measure is used to
			compute the distance between two sequences.
		"""
		return self._timesCalled

	# CONSTRUCTOR #

	def __init__(self) -> None:
		"""
			This method initializes the distance measure by setting the counter on the amount of times that the distance
			measure is used to compute the distance between two sequences to zero.
		"""
		self._timesCalled = 0

	# PUBLIC METHODS #

	def calculateDistance(self, sequenceOne: ndarray, sequenceTwo: ndarray) -> float:
		"""
			This method calculates the distance between two sequences.

			:param sequenceOne: The first sequence for which the distance should be computed.
			:param sequenceTwo: The second sequence for which the distance should be computed.
			:return: The distance between the two sequences.
		"""
		self._timesCalled += 1
		return fastdtw(sequenceOne, sequenceTwo, dist=euclidean)[0]

	def reset(self) -> None:
		"""
			This method resets the counter on the amount of times that the distance measure is used to compute the
			distance between two sequences.

			:return: void
		"""
		self._timesCalled = 0



"""
    This module contains a class called 'SequenceStore' that stores the ambiguous sequence that have yet to be clustered
     in the 'SeqClu' algorithm.
"""

from typing import Dict

from numpy import ndarray

from seqclupv.library.utilities.hash_sequence import hashSequence


class SequenceStore:

    # PROPERTIES #

    @property
    def sequences(self) -> Dict[str, ndarray]:
        """
            This method is a property that returns the stored set of ambiguous sequenceStore. The sequenceStore are stored in a
            dictionary where key is the hash of the sequence and the value is the sequence itself.

            :return: A dictionary where key is the hash of the sequence and the value is the sequence itself.
        """
        return self._sequences

    @property
    def sequenceHistory(self) -> Dict[str, int]:
        """
            This method is a property that returns a dictionary where the key is the hash of some sequence and the
            value is the tick at which the sequence was added to the buffer.

            :return: A dictionary where the key is the hash of some sequence and the
            value is the tick at which the sequence was added to the buffer.
        """
        return self._sequenceHistory

    @property
    def lastUpdate(self) -> int:
        """
            This method is a property that returns the tick at which the buffer of sequenceStore was last updated.

            :return: The tick at which the buffer of sequenceStore was last updated.
        """
        return self._lastUpdate

    # CONSTRUCTOR #

    def __init__(self, tick: int) -> None:
        """
            This method initializes the buffer of sequenceStore with a given tick at which the buffer is
            initialized.

            :param tick: The tick at which the buffer of sequenceStore is initialized.
        """
        assert tick >= -1
        self._sequences = {}
        self._sequenceHistory = {}
        self._lastUpdate = tick

    # PUBLIC METHODS #

    def addToSequences(self, sequence: ndarray, tick: int) -> None:
        """
            This method adds an incoming sequence to the buffer of sequenceStore.

            :param sequence: The incoming sequence that needs to be added to the buffer of sequenceStore.
            :param tick: The tick at which the incoming sequence is promoted to a sequence.
            :return: void
        """
        sequenceHash = hashSequence(sequence)
        assert sequenceHash not in self.sequences

        self._sequences[sequenceHash] = sequence
        self._sequenceHistory[sequenceHash] = tick
        self._lastUpdate = tick

    def getSequence(self, sequenceHash: str) -> ndarray:
        """
            This method returns a sequence given its hash.

            :param sequenceHash: The hash of the sequence that is requested.
            :return: The requested sequence.
        """
        # A requested sequence identified by its hash must always be available in the 'sequenceStore' data structure.
        assert sequenceHash in self.sequences
        return self.sequences[sequenceHash]

    def lastUpdateSequence(self, sequenceHash: str) -> int:
        """
            This method returns the tick at which a sequence that is stored in the buffer was last updated.

            :param sequenceHash: The hash of the sequence for which the information about when the sequence was last
            updated is requested.
            :return: The tick at which a sequence that is stored in the buffer was last updated.
        """
        # Since the data structure will only keep track of current sequenceStore and this function is only intended for use
        # with sequenceStore, the constraint that the provided hash of the sequenceStore needs to be contained in the data
        # structure that stores the ticks at which sequenceStore were updated is added.
        assert sequenceHash in self.sequenceHistory
        return self.sequenceHistory[sequenceHash]

    def removeFromSequences(self, sequenceHash: str) -> None:
        """
            This method removes a sequence from the set of sequenceStore. This occurs when the sequence is fully processed
            and discarded.

            :param sequenceHash: The hash of the sequence.
            :return: void
        """
        assert sequenceHash in self.sequences

        del self.sequences[sequenceHash]
        del self.sequenceHistory[sequenceHash]



"""
    This module contains a class called 'PrototypeStore' that is responsible for storing the prototypes of one of the
    clusters that is obtained as a result of executing the 'SeqClu' algorithm.
"""

from typing import Dict, Optional, Set, Tuple

from numpy import ndarray


class PrototypeStore:

    # PROPERTIES #

    @property
    def fullyInitialized(self) -> bool:
        """
            This method is a property that returns a boolean indicating whether or not the data structure has been
            initialized in full.

            :return: A boolean indicating whether or not the data structure has been
            initialized in full.
        """
        return self._fullyInitialized

    @property
    def lastUpdate(self) -> int:
        """
            This method is a property that returns the tick at which the data structure of prototypes was last
            updated.

            :return: The tick at which the data structure of prototypes was last updated.
        """
        assert self._lastUpdate >= 0
        return self._lastUpdate

    @property
    def numRepresentativePrototypes(self) -> int:
        """
            This method is a property that returns the number of representative prototypes of some cluster.

            :return: The number of representative prototypes of the cluster.
        """
        assert not self.fullyInitialized or self.updatingPrototypes or self._numRepresentativePrototypes <= self._numPrototypes
        return self._numRepresentativePrototypes

    @property
    def numOtherPrototypes(self) -> int:
        """
            This method is a property that returns the number of prototypes that are not representative prototypes
            of some cluster.

            :return: The number of prototypes that are not representative prototypes of some cluster.
        """
        return self._numPrototypes - self._numRepresentativePrototypes

    @property
    def numPrototypes(self) -> int:
        """
            This method is a property that returns the total number of prototypes of some cluster.

            :return: The total number of prototypes of some cluster.
        """
        assert not self.fullyInitialized or self.updatingPrototypes or self._numPrototypes >= self._numRepresentativePrototypes
        return self._numPrototypes

    @property
    def otherPrototypeHashes(self) -> Set[str]:
        """
            This method is a property that returns the hashes of the prototypes that are not representative prototypes
            of some cluster.

            :return: The hashes of the prototypes that are not representative prototypes of some cluster.
        """
        assert not self.fullyInitialized or self.updatingPrototypes or len(self._otherPrototypeHashes) == self.numOtherPrototypes
        return self._otherPrototypeHashes

    @property
    def prototypeHistory(self) -> Dict[str, int]:
        """
            This method is a property that returns a dictionary where the key is the hash of some prototype and the
            value is the tick at which the prototype was last updated.

            :return: A dictionary where the key is the hash of some prototype and the value is the tick at which
            the prototype was last updated.
        """
        return self._prototypeHistory

    @property
    def prototypes(self) -> Dict[str, ndarray]:
        """
            This method is a property that returns a dictionary where the key is the hash of some prototype and the
            value is the prototype itself.

            :return: A dictionary where the key is the hash of some prototype and the value is the prototype itself.
        """
        assert not self.fullyInitialized or self.updatingPrototypes or len(self._prototypes) == self._numPrototypes
        return self._prototypes

    @property
    def representativePrototypesInitialized(self) -> bool:
        """
            This property stores a boolean value indicating whether or not the representative prototypes have been
            initialized.

            :return: A boolean value indicating whether or not the representative prototypes have been initialized.
        """
        return self._representativePrototypesInitialized

    @property
    def representativePrototypeHashes(self) -> Set[str]:
        """
            This method is a property that returns the hashes of the prototypes that are representative prototypes
            of some cluster.

            :return: The hashes of the prototypes that are representative prototypes of some cluster.
        """
        assert not self.fullyInitialized or self.updatingPrototypes or len(self._representativePrototypeHashes) == self._numRepresentativePrototypes
        return self._representativePrototypeHashes

    @property
    def updatingPrototypes(self) -> bool:
        """
            This property stores a boolean value indicating whether or not the prototypes are currently being updated.

            :return: A boolean value indicating whether or not the prototypes are currently being updated.
        """
        return self._updatingPrototypes

    # CONSTRUCTOR #

    def __init__(self, numRepresentativePrototypes: int, numPrototypes: int, tick: int) -> None:
        """
            This method initializes the 'PrototypeStore' class with a given number of representative prototypes, number
            of total prototypes and the tick at which the class is initialized.

            :param numRepresentativePrototypes: The number of representative prototypes.
            :param numPrototypes: The total number of prototypes.
            :param tick: The tick at which the class is initialized.
            :return: void
        """
        assert 0 < numRepresentativePrototypes < numPrototypes and numPrototypes > 0 and tick >= -1
        self._fullyInitialized = False
        self._lastUpdate = tick
        self._numRepresentativePrototypes = numRepresentativePrototypes
        self._numPrototypes = numPrototypes
        self._otherPrototypeHashes = set([])
        self._prototypeHistory = {}
        self._prototypes = {}
        self._representativePrototypesInitialized = False
        self._representativePrototypeHashes = set([])
        self._updatingPrototypes = False

    # PUBLIC METHODS #

    def addPrototype(self, prototype: Tuple[Optional[str], Optional[ndarray]], representative: bool,
                     tick: int, toReplaceHash: Optional[str]) -> None:
        """
            This method adds a prototype to the set of prototypes that is stored in this class.

            :param prototype: The sequence that needs to be added as prototype.
            :param representative: A boolean value indicating whether or not the sequence should become a representative
            prototype.
            :param tick: The tick at which the sequence is added to the set of prototypes that is stored in this class.
            :param toReplaceHash: An optional hash of the sequence that needs to be demoted to a regular sequence from
            a prototype.
            :return: void
        """
        prototypeHash, prototypeData = prototype
        # This method should never be called if the cluster prototypes has been initialized in full.
        assert (not self.fullyInitialized and toReplaceHash is None) or \
               (toReplaceHash is not None and toReplaceHash in self.prototypes)

        if toReplaceHash is not None:
            self._removePrototype(toReplaceHash, representative)
        # The private 'addPrototype' method can now be called.
        self._addPrototype(prototypeHash, representative, tick)
        # The private 'addPrototype' method terminated successfully, hence the prototype can be stored.
        self._prototypes[prototypeHash] = prototypeData

        # After a new prototype is added, check if the data structures storing the prototypes contain 'numPrototypes'
        # prototypes. If so, the prototypes is fully initialized and constraints on the prototypes are put in effect.
        if len(self.representativePrototypeHashes) == self.numRepresentativePrototypes:
            print(
                f"[SeqClu] Prototypes have been representative initialized at tick {tick} with {len(self.representativePrototypeHashes)} "
                f"and {len(self.otherPrototypeHashes)}")
            self._representativePrototypesInitialized = True
        if len(self.representativePrototypeHashes) + len(self.otherPrototypeHashes) == self.numPrototypes:
            print(f"[SeqClu] Prototypes have been fully initialized at tick {tick} with {len(self.representativePrototypeHashes)} "
                  f"and {len(self.otherPrototypeHashes)}")
            self._fullyInitialized = True

    def getPrototype(self, prototypeHash: str) -> ndarray:
        """
            This method returns a prototype given the hash of this prototype.

            :param prototypeHash: The hash of the prototype that is requested.
            :return: The requested prototype.
        """
        # A requested prototype identified by its hash must always be available in the 'prototypes' data structure.
        assert prototypeHash in self.prototypes
        return self.prototypes[prototypeHash]

    def lastUpdatePrototype(self, prototypeHash: str) -> Optional[int]:
        """
            This method returns the tick at which some prototype was last updated given the hash of this prototype.

            :param prototypeHash: The hash of the prototype for which information about when it was last updated
            is requested.
            :return: The tick at which the prototype corresponding to the given hash was last updated.
        """
        # Since the data structure will only keep track of current prototypes and this method is only intended for use
        # with prototypes, the constraint that the provided hash of the prototype needs to be contained in the data
        # structure that stores the ticks at which prototypes were updated is added.
        assert prototypeHash in self.prototypeHistory
        return self.prototypeHistory[prototypeHash]

    def updatePrototypes(self, newPrototypes: Dict[str, ndarray], newOtherPrototypeHashes: Set[str],
                         newRepresentativePrototypeHashes: Set[str], tick: int) -> Set[str]:
        """
            This method updates the prototypes stored in the prototype store.

            :param newPrototypes: The new prototypes as a dictionary where the keys are the hashes of the new prototypes
            and the values are the prototypes itself.
            :param newOtherPrototypeHashes: The hashes of new non-representative prototypes.
            :param newRepresentativePrototypeHashes: The hashes of new representative prototypes.
            :param tick: The tick at which the prototypes are updated.
            :return: The hashes of the prototypes that were removed.
        """
        assert len(newPrototypes) == self.numPrototypes
        assert len(newOtherPrototypeHashes) == self.numOtherPrototypes
        assert len(newRepresentativePrototypeHashes) == self.numRepresentativePrototypes
        assert set(newPrototypes.keys()) == newOtherPrototypeHashes.union(newRepresentativePrototypeHashes)
        assert newOtherPrototypeHashes.intersection(newRepresentativePrototypeHashes) == set([])

        self._updatingPrototypes = True
        removedPrototypeHashes: Set[str] = set([])

        for prototypeHash in self.prototypes.keys():
            if prototypeHash not in newPrototypes:
                removedPrototypeHashes.add(prototypeHash)
                if prototypeHash in self.representativePrototypeHashes:
                    self._removePrototype(prototypeHash, True)
                else:
                    self._removePrototype(prototypeHash, False)

        for prototypeHash in self.representativePrototypeHashes:
            if prototypeHash in newOtherPrototypeHashes:
                self.prototypeHistory[prototypeHash] = tick

        for prototypeHash in self.otherPrototypeHashes:
            if prototypeHash in newRepresentativePrototypeHashes:
                self.prototypeHistory[prototypeHash] = tick

        self._prototypes = newPrototypes
        self._representativePrototypeHashes = newRepresentativePrototypeHashes
        self._otherPrototypeHashes = newOtherPrototypeHashes

        for otherPrototypeHash in newOtherPrototypeHashes:
            if otherPrototypeHash not in self.prototypeHistory:
                self.prototypeHistory[otherPrototypeHash] = tick

        for representativePrototypeHash in newRepresentativePrototypeHashes:
            if representativePrototypeHash not in self.prototypeHistory:
                self.prototypeHistory[representativePrototypeHash] = tick

        self._updatingPrototypes = False
        self._lastUpdate = tick
        _ = self.numOtherPrototypes
        _ = self.numRepresentativePrototypes
        _ = self.numPrototypes

        return removedPrototypeHashes

    # PRIVATE METHODS #

    def _addPrototype(self, prototypeHash: str, representative: bool, tick: int) -> None:
        """
            This method adds a prototype to the set of prototypes that is stored in this class.

            :param prototypeHash: The hash of the sequence that needs to be added as prototype.
            :param representative: A boolean value indicating whether or not the sequence should become a representative
            prototype.
            :param tick: The tick at which the sequence is added to the set of prototypes that is stored in this class.
            :return: void
        """
        # The sequence that needs to become a prototype can never already be a prototype.
        assert prototypeHash not in self.representativePrototypeHashes and prototypeHash not in self.otherPrototypeHashes

        if representative:
            assert len(self.representativePrototypeHashes) < self.numRepresentativePrototypes
            self.representativePrototypeHashes.add(prototypeHash)
        else:
            assert len(self.otherPrototypeHashes) < self.numOtherPrototypes
            self.otherPrototypeHashes.add(prototypeHash)
        self.prototypeHistory[prototypeHash] = tick
        self._lastUpdate = tick

    def _removePrototype(self, prototypeHash: str, representative: bool) -> None:
        """
            This method removes a prototype from the set of prototypes that is stored in this class.

            :param prototypeHash: The hash of the sequence that needs to be removed from the set of prototypes.
            :param representative: A boolean value indicating whether or not the sequence is a representative prototype.
            :return: void
        """
        if representative:
            assert prototypeHash in self.representativePrototypeHashes
            self.representativePrototypeHashes.remove(prototypeHash)
        else:
            assert prototypeHash in self.otherPrototypeHashes
            self.otherPrototypeHashes.remove(prototypeHash)
        del self.prototypeHistory[prototypeHash]



"""
    This package contains implementations of distance measures that are used in the 'SeqClu' algorithm to compute the
    distance between two incoming data points.
"""

from .dynamic_time_warping import DynamicTimeWarping


__all__ = ["DynamicTimeWarping"]
